import hashlib
import json
import logging
import math
import multiprocessing
import os
import pickle
import random
import string
import tempfile
from datetime import datetime, timedelta
from pathlib import Path
from time import sleep
from typing import (
    Any,
    cast,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)
from urllib.parse import urlencode, urlparse

import requests
import semver
from gi.repository import Gtk

from sublime_music.util import resolve_path

from .api_objects import Directory, Response
from .. import (
    Adapter,
    AlbumSearchQuery,
    api_objects as API,
    ConfigParamDescriptor,
    ConfigurationStore,
    ConfigureServerForm,
    UIInfo,
)

try:
    import gi

    gi.require_version("NM", "1.0")
    from gi.repository import NM

    networkmanager_imported = True
except Exception:
    # I really don't care what kind of exception it is, all that matters is the
    # import failed for some reason.
    networkmanager_imported = False

REQUEST_DELAY: Optional[Tuple[float, float]] = None
if delay_str := os.environ.get("REQUEST_DELAY"):
    if "," in delay_str:
        high, low = map(float, delay_str.split(","))
        REQUEST_DELAY = (high, low)
    else:
        REQUEST_DELAY = (float(delay_str), float(delay_str))

NETWORK_ALWAYS_ERROR: bool = False
if always_error := os.environ.get("NETWORK_ALWAYS_ERROR"):
    NETWORK_ALWAYS_ERROR = True


class ServerError(Exception):
    def __init__(self, status_code: int, message: str):
        self.status_code = status_code
        super().__init__(message)


class SubsonicAdapter(Adapter):
    """
    Defines an adapter which retrieves its data from a Subsonic server
    """

    # Configuration and Initialization Properties
    # ==================================================================================
    @staticmethod
    def get_ui_info() -> UIInfo:
        return UIInfo(
            name="Subsonic",
            description="Connect to a Subsonic-compatible server",
            icon_basename="subsonic",
            icon_dir=resolve_path("adapters/subsonic/icons"),
        )

    @staticmethod
    def get_configuration_form(config_store: ConfigurationStore) -> Gtk.Box:
        configs = {
            "server_address": ConfigParamDescriptor(str, "Server Address"),
            "username": ConfigParamDescriptor(str, "Username"),
            "password": ConfigParamDescriptor("password", "Password"),
            "verify_cert": ConfigParamDescriptor(
                bool,
                "Verify Certificate",
                default=True,
                advanced=True,
                helptext="Whether or not to verify the SSL certificate of the server.",
            ),
            "sync_enabled": ConfigParamDescriptor(
                bool,
                "Sync Play Queue",
                default=True,
                advanced=True,
                helptext="If toggled, Sublime Music will periodically save the play "
                "queue state so that you can resume on other devices.",
            ),
            "salt_auth": ConfigParamDescriptor(
                bool,
                "Use Salt Authentication",
                default=True,
                advanced=True,
                helptext="If toggled, Sublime Music will use salted hash tokens "
                "instead of the plain password in the request urls (only supported on "
                "Subsonic API 1.13.0+)",
            ),
        }

        if networkmanager_imported:
            configs.update(
                {
                    "local_network_ssid": ConfigParamDescriptor(
                        str,
                        "Local Network SSID",
                        advanced=True,
                        required=False,
                        helptext="If Sublime Music is connected to the given SSID, the "
                        "Local Network Address will be used instead of the Server "
                        "address when making network requests.",
                    ),
                    "local_network_address": ConfigParamDescriptor(
                        str,
                        "Local Network Address",
                        advanced=True,
                        required=False,
                        helptext="If Sublime Music is connected to the given Local "
                        "Network SSID, this URL will be used instead of the Server "
                        "address when making network requests.",
                    ),
                }
            )

        def verify_configuration() -> Dict[str, Optional[str]]:
            errors: Dict[str, Optional[str]] = {}

            with tempfile.TemporaryDirectory() as tmp_dir_name:
                try:
                    tmp_adapter = SubsonicAdapter(config_store, Path(tmp_dir_name))
                    tmp_adapter._get_json(
                        tmp_adapter._make_url("ping"),
                        timeout=2,
                        is_exponential_backoff_ping=True,
                    )
                except requests.exceptions.SSLError:
                    errors["__ping__"] = (
                        "<b>Error connecting to the server.</b>\n"
                        "An SSL error occurred while connecting to the server.\n"
                        "You may need to explicitly specify http://."
                    )
                except requests.ConnectionError:
                    errors["__ping__"] = (
                        "<b>Unable to connect to the server.</b>\n"
                        "Double check the server address."
                    )
                except ServerError as e:
                    if e.status_code in (10, 41) and config_store["salt_auth"]:
                        # status code 10: if salt auth is not enabled, server will
                        #   return error server error with status_code 10 since it'll
                        #   interpret it as a missing (password) parameter
                        # status code 41: as per subsonic api docs, description of
                        #   status_code 41 is "Token authentication not supported for
                        #   LDAP users." so fall back to password auth
                        try:
                            config_store["salt_auth"] = False
                            tmp_adapter = SubsonicAdapter(
                                config_store, Path(tmp_dir_name)
                            )
                            tmp_adapter._get_json(
                                tmp_adapter._make_url("ping"),
                                timeout=2,
                                is_exponential_backoff_ping=True,
                            )
                            logging.warn(
                                "Salted auth not supported, falling back to regular "
                                "password auth"
                            )
                        except ServerError as retry_e:
                            config_store["salt_auth"] = True
                            errors["__ping__"] = (
                                "<b>Error connecting to the server.</b>\n"
                                f"Error {retry_e.status_code}: {str(retry_e)}"
                            )
                    else:
                        errors["__ping__"] = (
                            "<b>Error connecting to the server.</b>\n"
                            f"Error {e.status_code}: {str(e)}"
                        )
                except Exception as e:
                    errors["__ping__"] = str(e)

            return errors

        return ConfigureServerForm(config_store, configs, verify_configuration)

    @staticmethod
    def migrate_configuration(config_store: ConfigurationStore):
        if "salt_auth" not in config_store:
            config_store["salt_auth"] = True

    def __init__(self, config: ConfigurationStore, data_directory: Path):
        self.data_directory = data_directory
        self.ignored_articles_cache_file = self.data_directory.joinpath(
            "ignored_articles.pickle"
        )

        self.hostname = config["server_address"]
        if (
            (ssid := config.get("local_network_ssid"))
            and (lan_address := config.get("local_network_address"))
            and networkmanager_imported
        ):
            networkmanager_client = NM.Client.new()

            # Only look at the active WiFi connections.
            for ac in networkmanager_client.get_active_connections():
                if ac.get_connection_type() != "802-11-wireless":
                    continue
                devs = ac.get_devices()
                if len(devs) != 1:
                    continue
                if devs[0].get_device_type() != NM.DeviceType.WIFI:
                    continue

                # If connected to the Local Network SSID, then change the hostname to
                # the Local Network Address.
                if ssid == ac.get_id():
                    self.hostname = lan_address
                    break

        parsed_hostname = urlparse(self.hostname)
        if not parsed_hostname.scheme:
            self.hostname = "https://" + self.hostname

        self.username = config["username"]
        self.password = cast(str, config.get_secret("password"))
        self.verify_cert = config["verify_cert"]
        self.use_salt_auth = config["salt_auth"]

        self.is_shutting_down = False
        self._ping_process: Optional[multiprocessing.Process] = None
        self._version = multiprocessing.Array("c", 20)
        self._offline_mode = False

        # TODO (#112): support XML?

    def initial_sync(self):
        # Try to ping the server five times using exponential backoff (2^5 = 32s).
        self._exponential_backoff(5)

    def shutdown(self):
        if self._ping_process:
            self._ping_process.terminate()

    # Availability Properties
    # ==================================================================================
    _server_available = multiprocessing.Value("b", False)
    _last_ping_timestamp = multiprocessing.Value("d", 0.0)

    def _exponential_backoff(self, n: int):
        logging.info(f"Starting Exponential Backoff: n={n}")
        if self._ping_process:
            self._ping_process.terminate()

        self._ping_process = multiprocessing.Process(
            target=self._check_ping_thread, args=(n,)
        )
        self._ping_process.start()

    def _check_ping_thread(self, n: int):
        i = 0
        while i < n and not self._offline_mode and not self._server_available.value:
            try:
                self._set_ping_status(timeout=2 * (i + 1))
            except Exception:
                pass
            sleep(2 ** i)
            i += 1

    def _set_ping_status(self, timeout: int = 2):
        logging.info(f"SET PING STATUS timeout={timeout}")
        now = datetime.now().timestamp()
        if now - self._last_ping_timestamp.value < 15:
            return

        # Try to ping the server.
        self._get_json(
            self._make_url("ping"),
            timeout=timeout,
            is_exponential_backoff_ping=True,
        )

    def on_offline_mode_change(self, offline_mode: bool):
        self._offline_mode = offline_mode

    @property
    def ping_status(self) -> bool:
        return self._server_available.value

    can_create_playlist = True
    can_delete_playlist = True
    can_get_album = True
    can_get_albums = True
    can_get_artist = True
    can_get_artists = True
    can_get_cover_art_uri = True
    can_get_directory = True
    can_get_ignored_articles = True
    can_get_playlist_details = True
    can_get_playlists = True
    can_get_song_details = True
    can_get_song_file_uri = True
    can_get_song_stream_uri = True
    can_scrobble_song = True
    can_search = True
    can_stream = True
    can_update_playlist = True

    def version_at_least(self, version: str) -> bool:
        if not self._version.value:
            return False
        return semver.VersionInfo.parse(self._version.value.decode()) >= version

    @property
    def can_get_genres(self) -> bool:
        return self.version_at_least("1.9.0")

    @property
    def can_get_play_queue(self) -> bool:
        return self.version_at_least("1.12.0")

    @property
    def can_save_play_queue(self) -> bool:
        return self.version_at_least("1.12.0")

    _schemes = None

    @property
    def supported_schemes(self) -> Iterable[str]:
        if not self._schemes:
            self._schemes = (urlparse(self.hostname)[0],)
        return self._schemes

    @property
    def supported_artist_query_types(self) -> Set[AlbumSearchQuery.Type]:
        supported = {
            AlbumSearchQuery.Type.RANDOM,
            AlbumSearchQuery.Type.NEWEST,
            AlbumSearchQuery.Type.FREQUENT,
            AlbumSearchQuery.Type.RECENT,
            AlbumSearchQuery.Type.STARRED,
            AlbumSearchQuery.Type.ALPHABETICAL_BY_NAME,
            AlbumSearchQuery.Type.ALPHABETICAL_BY_ARTIST,
        }
        if self.version_at_least("1.10.1"):
            supported.add(AlbumSearchQuery.Type.YEAR_RANGE)
            supported.add(AlbumSearchQuery.Type.GENRE)

        return supported

    # Helper mothods for making requests
    # ==================================================================================
    def _get_params(self) -> Dict[str, str]:
        """
        Gets the parameters that are needed for all requests to the Subsonic API. See
        Subsonic API Introduction for details.
        """
        params = {
            "u": self.username,
            "c": "Sublime Music",
            "f": "json",
            "v": self._version.value.decode() or "1.8.0",
        }

        if self.use_salt_auth:
            salt, token = self._generate_auth_token()
            params["s"] = salt
            params["t"] = token
        else:
            params["p"] = self.password

        return params

    def _generate_auth_token(self) -> Tuple[str, str]:
        """
        Generates the necessary authentication data to call the Subsonic API See the
        Authentication section of www.subsonic.org/pages/api.jsp for more information
        """
        salt = "".join(random.choices(string.ascii_letters + string.digits, k=8))
        token = hashlib.md5(f"{self.password}{salt}".encode()).hexdigest()
        return (salt, token)

    def _make_url(self, endpoint: str) -> str:
        return f"{self.hostname}/rest/{endpoint}.view"

    # TODO (#196) figure out some way of rate limiting requests. They often come in too
    # fast.
    def _get(
        self,
        url: str,
        timeout: Union[float, Tuple[float, float], None] = None,
        is_exponential_backoff_ping: bool = False,
        **params,
    ) -> Any:
        params = {**self._get_params(), **params}
        logging.info(f"[START] get: {url}")

        try:
            if REQUEST_DELAY is not None:
                delay = random.uniform(*REQUEST_DELAY)
                logging.info(f"REQUEST_DELAY enabled. Pausing for {delay} seconds")
                sleep(delay)
                if timeout:
                    if type(timeout) == tuple:
                        if delay > cast(Tuple[float, float], timeout)[0]:
                            raise TimeoutError("DUMMY TIMEOUT ERROR")
                    else:
                        if delay > cast(float, timeout):
                            raise TimeoutError("DUMMY TIMEOUT ERROR")

            if NETWORK_ALWAYS_ERROR:
                raise ServerError(69, "NETWORK_ALWAYS_ERROR enabled")

            # Deal with datetime parameters (convert to milliseconds since 1970)
            for k, v in params.items():
                if isinstance(v, datetime):
                    params[k] = int(v.timestamp() * 1000)

            if self._is_mock:
                logging.info("Using mock data")
                result = self._get_mock_data()
            else:
                result = requests.get(
                    url,
                    params=params,
                    verify=self.verify_cert,
                    timeout=timeout,
                )

            if result.status_code != 200:
                raise ServerError(
                    result.status_code, f"{url} returned status={result.status_code}."
                )

            # Any time that a server request succeeds, then we win.
            self._server_available.value = True
            self._last_ping_timestamp.value = datetime.now().timestamp()

        except Exception:
            logging.exception(f"[FAIL] get: {url} failed")
            self._server_available.value = False
            self._last_ping_timestamp.value = datetime.now().timestamp()
            if not is_exponential_backoff_ping:
                self._exponential_backoff(5)
            raise

        logging.info(f"[FINISH] get: {url}")
        return result

    def _get_json(
        self,
        url: str,
        timeout: Union[float, Tuple[float, float], None] = None,
        is_exponential_backoff_ping: bool = False,
        **params: Union[None, str, datetime, int, Sequence[int], Sequence[str]],
    ) -> Response:
        """
        Make a get request to a *Sonic REST API. Handle all types of errors including
        *Sonic ``<error>`` responses.

        :returns: a dictionary of the subsonic response.
        :raises Exception: needs some work
        """
        result = self._get(
            url,
            timeout=timeout,
            is_exponential_backoff_ping=is_exponential_backoff_ping,
            **params,
        )
        subsonic_response = result.json().get("subsonic-response")

        if not subsonic_response:
            raise ServerError(500, f"{url} returned invalid JSON.")

        if subsonic_response["status"] != "ok":
            raise ServerError(
                subsonic_response["error"].get("code"),
                subsonic_response["error"].get("message"),
            )

        self._version.value = subsonic_response["version"].encode()

        logging.debug(f"Response from {url}: {subsonic_response}")
        return Response.from_dict(subsonic_response)

    # Helper Methods for Testing
    _get_mock_data: Any = None
    _is_mock: bool = False

    def _set_mock_data(self, data: Any):
        class MockResult:
            status_code = 200

            def __init__(self, content: Any):
                self._content = content

            def content(self) -> Any:
                return self._content

            def json(self) -> Any:
                return json.loads(self._content)

        def get_mock_data() -> Any:
            if type(data) == Exception:
                raise data
            if hasattr(data, "__next__"):
                if d := next(data):
                    logging.info("MOCK DATA: %s", d)
                    return MockResult(d)

            logging.info("MOCK DATA: %s", data)
            return MockResult(data)

        self._get_mock_data = get_mock_data

    # Data Retrieval Methods
    # ==================================================================================
    def get_playlists(self) -> Sequence[API.Playlist]:
        if playlists := self._get_json(self._make_url("getPlaylists")).playlists:
            return sorted(playlists.playlist, key=lambda p: p.name.lower())
        return []

    def get_playlist_details(self, playlist_id: str) -> API.Playlist:
        result = self._get_json(self._make_url("getPlaylist"), id=playlist_id).playlist
        assert result, f"Error getting playlist {playlist_id}"
        return result

    def create_playlist(
        self, name: str, songs: Sequence[API.Song] = None
    ) -> Optional[API.Playlist]:
        return self._get_json(
            self._make_url("createPlaylist"),
            name=name,
            songId=[s.id for s in songs or []],
        ).playlist

    def update_playlist(
        self,
        playlist_id: str,
        name: str = None,
        comment: str = None,
        public: bool = None,
        song_ids: Sequence[str] = None,
        append_song_ids: Sequence[str] = None,
    ) -> API.Playlist:
        if any(x is not None for x in (name, comment, public, append_song_ids)):
            self._get_json(
                self._make_url("updatePlaylist"),
                playlistId=playlist_id,
                name=name,
                comment=comment,
                public=public,
                songIdToAdd=append_song_ids,
            )

        playlist = None
        if song_ids is not None:
            playlist = self._get_json(
                self._make_url("createPlaylist"),
                playlistId=playlist_id,
                songId=song_ids,
            ).playlist

        # If the call to createPlaylist to update the song IDs returned the playlist,
        # return it.
        return playlist or self.get_playlist_details(playlist_id)

    def delete_playlist(self, playlist_id: str):
        self._get_json(self._make_url("deletePlaylist"), id=playlist_id)

    def get_cover_art_uri(self, cover_art_id: str, scheme: str, size: int) -> str:
        params = {"id": cover_art_id, "size": size, **self._get_params()}
        return self._make_url("getCoverArt") + "?" + urlencode(params)

    def get_song_file_uri(self, song_id: str, schemes: Iterable[str]) -> str:
        assert any(s in schemes for s in self.supported_schemes)
        params = {"id": song_id, **self._get_params()}
        return self._make_url("download") + "?" + urlencode(params)

    def get_song_stream_uri(self, song_id: str) -> str:
        params = {"id": song_id, **self._get_params()}
        return self._make_url("stream") + "?" + urlencode(params)

    def get_song_details(self, song_id: str) -> API.Song:
        song = self._get_json(self._make_url("getSong"), id=song_id).song
        assert song, f"Error getting song {song_id}"
        return song

    def scrobble_song(self, song: API.Song):
        self._get(self._make_url("scrobble"), id=song.id)

    def get_artists(self) -> Sequence[API.Artist]:
        if artist_index := self._get_json(self._make_url("getArtists")).artists:
            with open(self.ignored_articles_cache_file, "wb+") as f:
                pickle.dump(artist_index.ignored_articles, f)

            artists = []
            for index in artist_index.index:
                artists.extend(index.artist)
            return cast(Sequence[API.Artist], artists)
        return []

    def get_artist(self, artist_id: str) -> API.Artist:
        artist = self._get_json(self._make_url("getArtist"), id=artist_id).artist
        assert artist, f"Error getting artist {artist_id}"
        if self.version_at_least("1.11.0"):
            try:
                artist_info = self._get_json(
                    self._make_url("getArtistInfo2"), id=artist_id
                )
                artist.augment_with_artist_info(artist_info.artist_info)
            except Exception:
                pass
        return artist

    def get_ignored_articles(self) -> Set[str]:
        ignored_articles = "The El La Los Las Le Les"
        try:
            # If we already got the ignored articles from the get_artists, do that here.
            with open(self.ignored_articles_cache_file, "rb+") as f:
                ignored_articles = pickle.load(f)
        except Exception:
            try:
                # Whatever the exception, fall back on getting from the server.
                if artists := self._get_json(self._make_url("getArtists")).artists:
                    ignored_articles = artists.ignored_articles
            except Exception:
                # Use the default ignored articles.
                pass

        return set(ignored_articles.split())

    def get_albums(
        self, query: AlbumSearchQuery, sort_direction: str = "ascending"
    ) -> Sequence[API.Album]:
        type_ = {
            AlbumSearchQuery.Type.RANDOM: "random",
            AlbumSearchQuery.Type.NEWEST: "newest",
            AlbumSearchQuery.Type.FREQUENT: "frequent",
            AlbumSearchQuery.Type.RECENT: "recent",
            AlbumSearchQuery.Type.STARRED: "starred",
            AlbumSearchQuery.Type.ALPHABETICAL_BY_NAME: "alphabeticalByName",
            AlbumSearchQuery.Type.ALPHABETICAL_BY_ARTIST: "alphabeticalByArtist",
            AlbumSearchQuery.Type.YEAR_RANGE: "byYear",
            AlbumSearchQuery.Type.GENRE: "byGenre",
        }[query.type]

        extra_args: Dict[str, Any] = {}
        if query.type == AlbumSearchQuery.Type.YEAR_RANGE:
            assert (year_range := query.year_range)
            extra_args = {
                "fromYear": year_range[0],
                "toYear": year_range[1],
            }
        elif query.type == AlbumSearchQuery.Type.GENRE:
            assert (genre := query.genre)
            extra_args = {"genre": genre.name}

        albums: List[API.Album] = []
        page_size = 50 if query.type == AlbumSearchQuery.Type.RANDOM else 500
        offset = 0

        def get_page(offset: int) -> Sequence[API.Album]:
            album_list = self._get_json(
                self._make_url("getAlbumList2"),
                type=type_,
                size=page_size,
                offset=offset,
                **extra_args,
            ).albums
            return album_list.album if album_list else []

        # Get all pages.
        while len(next_page := get_page(offset)) > 0:
            albums.extend(next_page)
            if query.type == AlbumSearchQuery.Type.RANDOM:
                break
            offset += page_size

        return albums

    def get_album(self, album_id: str) -> API.Album:
        album = self._get_json(self._make_url("getAlbum"), id=album_id).album
        assert album, f"Error getting album {album_id}"
        return album

    def _get_indexes(self) -> API.Directory:
        indexes = self._get_json(self._make_url("getIndexes")).indexes
        assert indexes, "Error getting indexes"
        with open(self.ignored_articles_cache_file, "wb+") as f:
            pickle.dump(indexes.ignored_articles, f)

        root_dir_items: List[Dict[str, Any]] = []
        for index in indexes.index:
            root_dir_items.extend(map(lambda x: {**x, "isDir": True}, index.artist))
        return Directory(id="root", _children=root_dir_items)

    def get_directory(self, directory_id: str) -> API.Directory:
        if directory_id == "root":
            return self._get_indexes()

        # TODO (#187) make sure to filter out all non-song files
        directory = self._get_json(
            self._make_url("getMusicDirectory"), id=directory_id
        ).directory
        assert directory, f"Error getting directory {directory_id}"
        return directory

    def get_genres(self) -> Sequence[API.Genre]:
        if genres := self._get_json(self._make_url("getGenres")).genres:
            return genres.genre
        return []

    def get_play_queue(self) -> Optional[API.PlayQueue]:
        return self._get_json(self._make_url("getPlayQueue")).play_queue

    def save_play_queue(
        self,
        song_ids: Sequence[str],
        current_song_index: int = None,
        position: timedelta = None,
    ):
        # TODO (sonic-extensions-api/specification#1) make an extension that allows you
        # to save the play queue position by index instead of id.
        self._get(
            self._make_url("savePlayQueue"),
            id=song_ids,
            timeout=2,
            current=song_ids[current_song_index]
            if current_song_index is not None
            else None,
            position=math.floor(position.total_seconds() * 1000) if position else None,
        )

    def search(self, query: str) -> API.SearchResult:
        result = self._get_json(self._make_url("search3"), query=query).search_result
        if not result:
            return API.SearchResult(query)

        search_result = API.SearchResult(query)
        search_result.add_results("albums", result.album)
        search_result.add_results("artists", result.artist)
        search_result.add_results("songs", result.song)
        return search_result



from .album_with_songs import AlbumWithSongs
from .icon_button import IconButton, IconMenuButton, IconToggleButton
from .load_error import LoadError
from .song_list_column import SongListColumn
from .spinner_image import SpinnerImage

__all__ = (
    "AlbumWithSongs",
    "IconButton",
    "IconMenuButton",
    "IconToggleButton",
    "LoadError",
    "SongListColumn",
    "SpinnerImage",
)



import abc
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from typing import Callable, Dict, Optional, Set, Tuple, Type, Union

from ..adapters.api_objects import Song


@dataclass
class PlayerEvent:
    """
    Represents an event triggered by the player. This is a way to signal state changes
    to Sublime Music if the player can be controlled outside of Sublime Music (for
    example, Chromecast player).

    Each player event has a :class:`PlayerEvent.EventType`. Additionally, each event
    type has additional information in the form of additional properties on the
    :class:`PlayerEvent` object.

    * :class:`PlayerEvent.EventType.PLAY_STATE_CHANGE` -- indicates that the play state
      of the player has changed. The :class:`PlayerEvent.playing` property is required
      for this event type.
    * :class:`PlayerEvent.EventType.VOLUME_CHANGE` -- indicates that the player's volume
      has changed. The :classs`PlayerEvent.volume` property is required for this event
      type and should be in the range [0, 100].
    * :class:`PlayerEvent.EventType.STREAM_CACHE_PROGRESS_CHANGE` -- indicates that the
      stream cache progress has changed. When streaming a song, this will be used to
      show how much of the song has been loaded into the player. The
      :class:`PlayerEvent.stream_cache_duration` property is required for this event
      type and should be a float represent the number of seconds of the song that have
      been cached.
    * :class:`PlayerEvent.EventType.CONNECTING` -- indicates that a device is being
      connected to. The :class:`PlayerEvent.device_id` property is required for this
      event type and indicates the device ID that is being connected to.
    * :class:`PlayerEvent.EventType.CONNECTED` -- indicates that a device has been
      connected to. The :class:`PlayerEvent.device_id` property is required for this
      event type and indicates the device ID that has been connected to.
    """

    class EventType(Enum):
        PLAY_STATE_CHANGE = 0
        VOLUME_CHANGE = 1
        STREAM_CACHE_PROGRESS_CHANGE = 2
        CONNECTING = 3
        CONNECTED = 4
        DISCONNECT = 5

    type: EventType
    device_id: str
    playing: Optional[bool] = None
    volume: Optional[float] = None
    stream_cache_duration: Optional[float] = None


@dataclass
class PlayerDeviceEvent:
    class Delta(Enum):
        ADD = 0
        REMOVE = 1

    delta: Delta
    player_type: Type
    id: str
    name: str


class Player(abc.ABC):
    song_loaded = False

    @property
    @abc.abstractmethod
    def enabled(self) -> bool:
        return True

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        :returns: returns the friendly name of the player for display in the UI.
        """

    @property
    @abc.abstractmethod
    def supported_schemes(self) -> Set[str]:
        """
        :returns: a set of all the schemes that the player can play.
        """

    @property
    def can_start_playing_with_no_latency(self) -> bool:
        """
        :returns: whether the player can start playing a song with no latency.
        """
        return False

    @staticmethod
    @abc.abstractmethod
    def get_configuration_options() -> Dict[str, Union[Type, Tuple[str, ...]]]:
        """
        :returns: a dictionary of configuration key -> type of the option or tuple of
            options (for a dropdown menu).
        """

    @abc.abstractmethod
    def __init__(
        self,
        on_timepos_change: Callable[[Optional[float]], None],
        on_track_end: Callable[[], None],
        on_player_event: Callable[[PlayerEvent], None],
        player_device_change_callback: Callable[[PlayerDeviceEvent], None],
        config: Dict[str, Union[str, int, bool]],
    ):
        """
        Initialize the player.

        :param config: A dictionary of configuration key -> configuration value.
        """

    @abc.abstractmethod
    def change_settings(self, config: Dict[str, Union[str, int, bool]]):
        """
        This function is called when the player settings are changed (normally this
        happens when the user changes the settings in the UI).

        :param config: A dictionary of configuration key -> configuration value.
        """

    @abc.abstractmethod
    def refresh_players(self):
        """
        This function is called when the user requests the player list to be refreshed
        in the UI.

        This function should call the ``player_device_change_callback`` with the delta
        events to indicate changes to the UI. If there is no reason to refresh (for
        example, the MPV player), then this function can do nothing.
        """

    @abc.abstractmethod
    def set_current_device_id(self, device_id: str):
        """
        Switch to the given device ID.
        """

    def reset(self):
        """
        Reset the player.
        """

    @abc.abstractmethod
    def shutdown(self):
        """
        Do any cleanup of the player.
        """

    @property
    @abc.abstractmethod
    def playing(self) -> bool:
        """
        :returns: whether or not the player is currently playing a song.
        """

    @abc.abstractmethod
    def get_volume(self) -> float:
        """
        :returns: the current volume on a scale of [0, 100]
        """

    @abc.abstractmethod
    def set_volume(self, volume: float):
        """
        Set the volume of the player to the given value.

        :param volume: the value to set the volume to. Will be in the range [0, 100]
        """

    @abc.abstractmethod
    def get_is_muted(self) -> bool:
        """
        :returns: whether or not the player is muted.
        """

    @abc.abstractmethod
    def set_muted(self, muted: bool):
        """
        :param muted: set the player's "muted" property to the given value.
        """

    @abc.abstractmethod
    def play_media(self, uri: str, progress: timedelta, song: Song):
        """
        :param uri: the URI to play. The URI is guaranteed to be one of the schemes in
            the :class:`supported_schemes` set for this adapter.
        :param progress: the time at which to start playing the song.
        :param song: the actual song. This could be used to set metadata and such on the
            player.
        """

    @abc.abstractmethod
    def pause(self):
        """
        Pause the player.
        """

    @abc.abstractmethod
    def play(self):
        """
        Play the current media.
        """

    def seek(self, position: timedelta):
        """
        :param position: seek to the given position in the song.
        """



import logging
import os
import random
import shutil
import sys
from datetime import timedelta
from functools import partial
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple
from urllib.parse import urlparse

import bleach

try:
    import osxmmkeys

    tap_imported = True
except Exception:
    tap_imported = False

from gi.repository import Gdk, GdkPixbuf, Gio, GLib, Gtk

try:
    import gi

    gi.require_version("Notify", "0.7")
    from gi.repository import Notify

    glib_notify_exists = True
except Exception:
    # I really don't care what kind of exception it is, all that matters is the
    # import failed for some reason.
    logging.warning(
        "Unable to import Notify from GLib. Notifications will be disabled."
    )
    glib_notify_exists = False

from .adapters import (
    AdapterManager,
    AlbumSearchQuery,
    CacheMissError,
    DownloadProgress,
    Result,
    SongCacheStatus,
)
from .adapters.api_objects import Playlist, PlayQueue, Song
from .config import AppConfiguration, ProviderConfiguration
from .dbus import dbus_propagate, DBusManager
from .players import PlayerDeviceEvent, PlayerEvent, PlayerManager
from .ui.configure_provider import ConfigureProviderDialog
from .ui.main import MainWindow
from .ui.state import RepeatType, UIState
from .util import resolve_path


class SublimeMusicApp(Gtk.Application):
    def __init__(self, config_file: Path):
        super().__init__(application_id="app.sublimemusic.SublimeMusic")
        if glib_notify_exists:
            Notify.init("Sublime Music")

        self.window: Optional[Gtk.Window] = None
        self.app_config = AppConfiguration.load_from_file(config_file)
        self.dbus_manager: Optional[DBusManager] = None

        self.connect("shutdown", self.on_app_shutdown)

    player_manager: Optional[PlayerManager] = None
    exiting: bool = False

    def do_startup(self):
        Gtk.Application.do_startup(self)

        def add_action(name: str, fn: Callable, parameter_type: str = None):
            """Registers an action with the application."""
            if type(parameter_type) == str:
                parameter_type = GLib.VariantType(parameter_type)
            action = Gio.SimpleAction.new(name, parameter_type)
            action.connect("activate", fn)
            self.add_action(action)

        # Add action for menu items.
        add_action("add-new-music-provider", self.on_add_new_music_provider)
        add_action("edit-current-music-provider", self.on_edit_current_music_provider)
        add_action(
            "switch-music-provider", self.on_switch_music_provider, parameter_type="s"
        )
        add_action(
            "remove-music-provider", self.on_remove_music_provider, parameter_type="s"
        )

        # Add actions for player controls
        add_action("play-pause", self.on_play_pause)
        add_action("next-track", self.on_next_track)
        add_action("prev-track", self.on_prev_track)
        add_action("repeat-press", self.on_repeat_press)
        add_action("shuffle-press", self.on_shuffle_press)

        # Navigation actions.
        add_action("play-next", self.on_play_next, parameter_type="as")
        add_action("add-to-queue", self.on_add_to_queue, parameter_type="as")
        add_action("go-to-album", self.on_go_to_album, parameter_type="s")
        add_action("go-to-artist", self.on_go_to_artist, parameter_type="s")
        add_action("browse-to", self.browse_to, parameter_type="s")
        add_action("go-to-playlist", self.on_go_to_playlist, parameter_type="s")

        add_action("go-online", self.on_go_online)
        add_action("refresh-devices", self.on_refresh_devices)
        add_action(
            "refresh-window",
            lambda *a: self.on_refresh_window(None, {}, True),
        )
        add_action("mute-toggle", self.on_mute_toggle)
        add_action(
            "update-play-queue-from-server",
            lambda a, p: self.update_play_state_from_server(),
        )

        if tap_imported:
            self.tap = osxmmkeys.Tap()
            self.tap.on("play_pause", self.on_play_pause)
            self.tap.on("next_track", self.on_next_track)
            self.tap.on("prev_track", self.on_prev_track)
            self.tap.start()

    def do_activate(self):
        # We only allow a single window and raise any existing ones
        if self.window:
            self.window.present()
            return

        # Configure Icons
        default_icon_theme = Gtk.IconTheme.get_default()
        for adapter in AdapterManager.available_adapters:
            if icon_dir := adapter.get_ui_info().icon_dir:
                default_icon_theme.append_search_path(str(icon_dir))

        icon_dirs = [resolve_path("ui/icons"), resolve_path("adapters/icons")]
        for icon_dir in icon_dirs:
            default_icon_theme.append_search_path(str(icon_dir))

        # Windows are associated with the application when the last one is
        # closed the application shuts down.
        self.window = MainWindow(application=self, title="Sublime Music")

        # Configure the CSS provider so that we can style elements on the
        # window.
        css_provider = Gtk.CssProvider()
        css_provider.load_from_path(str(resolve_path("ui/app_styles.css")))
        context = Gtk.StyleContext()
        screen = Gdk.Screen.get_default()
        context.add_provider_for_screen(
            screen, css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER
        )

        self.window.show_all()
        self.window.present()

        # Load the state for the server, if it exists.
        self.app_config.load_state()

        # If there is no current provider, use the first one if there are any
        # configured, and if none are configured, then show the dialog to create a new
        # one.
        if self.app_config.provider is None:
            if len(self.app_config.providers) == 0:
                self.show_configure_servers_dialog()

                # If they didn't add one with the dialog, close the window.
                if len(self.app_config.providers) == 0:
                    self.window.close()
                    return

        AdapterManager.reset(self.app_config, self.on_song_download_progress)

        # Connect after we know there's a server configured.
        self.window.stack.connect("notify::visible-child", self.on_stack_change)
        self.window.connect("song-clicked", self.on_song_clicked)
        self.window.connect("songs-removed", self.on_songs_removed)
        self.window.connect("refresh-window", self.on_refresh_window)
        self.window.connect("notification-closed", self.on_notification_closed)
        self.window.connect("go-to", self.on_window_go_to)
        self.window.connect("key-press-event", self.on_window_key_press)
        self.window.player_controls.connect("song-scrub", self.on_song_scrub)
        self.window.player_controls.connect("device-update", self.on_device_update)
        self.window.player_controls.connect("volume-change", self.on_volume_change)

        # Configure the players
        self.last_play_queue_update = timedelta(0)
        self.loading_state = False
        self.should_scrobble_song = False

        def on_timepos_change(value: Optional[float]):
            if (
                self.loading_state
                or not self.window
                or not self.app_config.state.current_song
            ):
                return

            if value is None:
                self.last_play_queue_update = timedelta(0)
                return

            self.app_config.state.song_progress = timedelta(seconds=value)
            GLib.idle_add(
                self.window.player_controls.update_scrubber,
                self.app_config.state.song_progress,
                self.app_config.state.current_song.duration,
                self.app_config.state.song_stream_cache_progress,
            )

            if (self.last_play_queue_update + timedelta(15)).total_seconds() <= value:
                self.save_play_queue()

            if (
                value > 5
                and self.should_scrobble_song
                and AdapterManager.can_scrobble_song()
            ):
                AdapterManager.scrobble_song(self.app_config.state.current_song)
                self.should_scrobble_song = False

        def on_track_end():
            at_end = (
                self.app_config.state.current_song_index
                == len(self.app_config.state.play_queue) - 1
            )
            no_repeat = self.app_config.state.repeat_type == RepeatType.NO_REPEAT
            if at_end and no_repeat:
                self.app_config.state.playing = False
                self.app_config.state.current_song_index = -1
                self.update_window()
                return

            GLib.idle_add(self.on_next_track)

        def on_player_event(event: PlayerEvent):
            if event.type == PlayerEvent.EventType.PLAY_STATE_CHANGE:
                assert event.playing is not None
                self.app_config.state.playing = event.playing
                if self.dbus_manager:
                    self.dbus_manager.property_diff()
                self.update_window()

            elif event.type == PlayerEvent.EventType.VOLUME_CHANGE:
                assert event.volume is not None
                self.app_config.state.volume = event.volume
                if self.dbus_manager:
                    self.dbus_manager.property_diff()
                self.update_window()

            elif event.type == PlayerEvent.EventType.STREAM_CACHE_PROGRESS_CHANGE:
                if (
                    self.loading_state
                    or not self.window
                    or not self.app_config.state.current_song
                    or event.stream_cache_duration is None
                ):
                    return
                self.app_config.state.song_stream_cache_progress = timedelta(
                    seconds=event.stream_cache_duration
                )
                GLib.idle_add(
                    self.window.player_controls.update_scrubber,
                    self.app_config.state.song_progress,
                    self.app_config.state.current_song.duration,
                    self.app_config.state.song_stream_cache_progress,
                )

            elif event.type == PlayerEvent.EventType.DISCONNECT:
                assert self.player_manager
                self.app_config.state.current_device = "this device"
                self.player_manager.set_current_device_id(
                    self.app_config.state.current_device
                )
                self.player_manager.set_volume(self.app_config.state.volume)
                self.update_window()

        def player_device_change_callback(event: PlayerDeviceEvent):
            assert self.player_manager
            state_device = self.app_config.state.current_device

            if event.delta == PlayerDeviceEvent.Delta.ADD:
                # If the device added is the one that's supposed to be active, activate
                # it and set the volume.
                if event.id == state_device:
                    self.player_manager.set_current_device_id(
                        self.app_config.state.current_device
                    )
                    self.player_manager.set_volume(self.app_config.state.volume)
                    self.app_config.state.connected_device_name = event.name
                    self.app_config.state.connecting_to_device = False
                self.app_config.state.available_players[event.player_type].add(
                    (event.id, event.name)
                )

            elif event.delta == PlayerDeviceEvent.Delta.REMOVE:
                if state_device == event.id:
                    self.player_manager.pause()
                self.app_config.state.available_players[event.player_type].remove(
                    (event.id, event.name)
                )

            self.update_window()

        self.app_config.state.connecting_to_device = True

        def check_if_connected():
            if not self.app_config.state.connecting_to_device:
                return
            self.app_config.state.current_device = "this device"
            self.app_config.state.connecting_to_device = False
            self.player_manager.set_current_device_id(
                self.app_config.state.current_device
            )
            self.update_window()

        self.player_manager = PlayerManager(
            on_timepos_change,
            on_track_end,
            on_player_event,
            lambda *a: GLib.idle_add(player_device_change_callback, *a),
            self.app_config.player_config,
        )
        GLib.timeout_add(10000, check_if_connected)

        # Update after Adapter Initial Sync
        def after_initial_sync(_):
            self.update_window()

            # Prompt to load the play queue from the server.
            if AdapterManager.can_get_play_queue():
                self.update_play_state_from_server(prompt_confirm=True)

            # Get the playlists, just so that we don't have tons of cache misses from
            # DBus trying to get the playlists.
            if AdapterManager.can_get_playlists():
                AdapterManager.get_playlists()

        inital_sync_result = AdapterManager.initial_sync()
        inital_sync_result.add_done_callback(after_initial_sync)

        # Send out to the bus that we exist.
        if self.dbus_manager:
            self.dbus_manager.property_diff()

    # ########## DBUS MANAGMENT ########## #
    def do_dbus_register(self, connection: Gio.DBusConnection, path: str) -> bool:
        self.dbus_manager = DBusManager(
            connection,
            self.on_dbus_method_call,
            self.on_dbus_set_property,
            lambda: (self.app_config, self.player_manager),
        )
        return True

    def on_dbus_method_call(
        self,
        connection: Gio.DBusConnection,
        sender: str,
        path: str,
        interface: str,
        method: str,
        params: GLib.Variant,
        invocation: Gio.DBusMethodInvocation,
    ):
        def seek_fn(offset: float):
            if not self.app_config.state.current_song:
                return
            new_seconds = self.app_config.state.song_progress + timedelta(
                microseconds=offset
            )

            # This should not ever happen. The current_song should always have
            # a duration, but the Child object has `duration` optional because
            # it could be a directory.
            assert self.app_config.state.current_song.duration is not None
            self.on_song_scrub(
                None,
                (
                    new_seconds.total_seconds()
                    / self.app_config.state.current_song.duration.total_seconds()
                )
                * 100,
            )

        def set_pos_fn(track_id: str, position: float = 0):
            if self.app_config.state.playing:
                self.on_play_pause()
            pos_seconds = timedelta(microseconds=position)
            self.app_config.state.song_progress = pos_seconds
            track_id, occurrence = track_id.split("/")[-2:]

            # Find the (N-1)th time that the track id shows up in the list. (N
            # is the -*** suffix on the track id.)
            song_index = [
                i
                for i, x in enumerate(self.app_config.state.play_queue)
                if x == track_id
            ][int(occurrence) or 0]

            self.play_song(song_index)

        def get_tracks_metadata(track_ids: List[str]) -> GLib.Variant:
            if not self.dbus_manager:
                return

            if len(track_ids) == 0:
                # We are lucky, just return an empty list.
                return GLib.Variant("(aa{sv})", ([],))

            # Have to calculate all of the metadatas so that we can deal with
            # repeat song IDs.
            metadatas: Iterable[Any] = [
                self.dbus_manager.get_mpris_metadata(
                    i,
                    self.app_config.state.play_queue,
                )
                for i in range(len(self.app_config.state.play_queue))
            ]

            # Get rid of all of the tracks that were not requested.
            metadatas = list(
                filter(lambda m: m["mpris:trackid"] in track_ids, metadatas)
            )

            assert len(metadatas) == len(track_ids)

            # Sort them so they get returned in the same order as they were
            # requested.
            metadatas = sorted(
                metadatas, key=lambda m: track_ids.index(m["mpris:trackid"])
            )

            # Turn them into dictionaries that can actually be serialized into
            # a GLib.Variant.
            metadatas = map(
                lambda m: {k: DBusManager.to_variant(v) for k, v in m.items()},
                metadatas,
            )

            return GLib.Variant("(aa{sv})", (list(metadatas),))

        def activate_playlist(playlist_id: str):
            playlist_id = playlist_id.split("/")[-1]
            playlist = AdapterManager.get_playlist_details(playlist_id).result()

            # Calculate the song id to play.
            song_idx = 0
            if self.app_config.state.shuffle_on:
                song_idx = random.randint(0, len(playlist.songs) - 1)

            self.on_song_clicked(
                None,
                song_idx,
                tuple(s.id for s in playlist.songs),
                {"active_playlist_id": playlist_id},
            )

        def get_playlists(
            index: int,
            max_count: int,
            order: str,
            reverse_order: bool,
        ) -> GLib.Variant:
            playlists_result = AdapterManager.get_playlists()
            if not playlists_result.data_is_available:
                # We don't want to wait for the response in this case, so just
                # return an empty array.
                return GLib.Variant("(a(oss))", ([],))

            playlists = list(playlists_result.result())

            sorters = {
                "Alphabetical": lambda p: p.name,
                "Created": lambda p: p.created,
                "Modified": lambda p: p.changed,
            }
            playlists.sort(
                key=sorters.get(order, lambda p: p),
                reverse=reverse_order,
            )

            def make_playlist_tuple(p: Playlist) -> GLib.Variant:
                cover_art_filename = AdapterManager.get_cover_art_uri(
                    p.cover_art,
                    "file",
                    allow_download=False,
                ).result()
                return (f"/playlist/{p.id}", p.name, cover_art_filename or "")

            return GLib.Variant(
                "(a(oss))",
                (
                    [
                        make_playlist_tuple(p)
                        for p in playlists[index : (index + max_count)]
                    ],
                ),
            )

        def play():
            if not self.app_config.state.playing:
                self.on_play_pause()

        def pause():
            if self.app_config.state.playing:
                self.on_play_pause()

        method_call_map: Dict[str, Dict[str, Any]] = {
            "org.mpris.MediaPlayer2": {
                "Raise": self.window and self.window.present,
                "Quit": self.window and self.window.destroy,
            },
            "org.mpris.MediaPlayer2.Player": {
                "Next": self.on_next_track,
                "Previous": self.on_prev_track,
                "Pause": pause,
                "PlayPause": self.on_play_pause,
                "Stop": pause,
                "Play": play,
                "Seek": seek_fn,
                "SetPosition": set_pos_fn,
            },
            "org.mpris.MediaPlayer2.TrackList": {
                "GoTo": set_pos_fn,
                "GetTracksMetadata": get_tracks_metadata,
                # 'RemoveTrack': remove_track,
            },
            "org.mpris.MediaPlayer2.Playlists": {
                "ActivatePlaylist": activate_playlist,
                "GetPlaylists": get_playlists,
            },
        }
        method_fn = method_call_map.get(interface, {}).get(method)
        if method_fn is None:
            logging.warning(f"Unknown/unimplemented method: {interface}.{method}.")
        invocation.return_value(method_fn(*params) if callable(method_fn) else None)

    def on_dbus_set_property(
        self,
        connection: Gio.DBusConnection,
        sender: str,
        path: str,
        interface: str,
        property_name: str,
        value: GLib.Variant,
    ):
        def change_loop(new_loop_status: GLib.Variant):
            self.app_config.state.repeat_type = RepeatType.from_mpris_loop_status(
                new_loop_status.get_string()
            )
            self.update_window()

        def set_shuffle(new_val: GLib.Variant):
            if new_val.get_boolean() != self.app_config.state.shuffle_on:
                self.on_shuffle_press(None, None)

        def set_volume(new_val: GLib.Variant):
            self.on_volume_change(None, new_val.get_double() * 100)

        setter_map: Dict[str, Dict[str, Any]] = {
            "org.mpris.MediaPlayer2.Player": {
                "LoopStatus": change_loop,
                "Rate": lambda _: None,
                "Shuffle": set_shuffle,
                "Volume": set_volume,
            }
        }

        setter = setter_map.get(interface, {}).get(property_name)
        if setter is None:
            logging.warning("Set: Unknown property: {property_name}.")
            return
        if callable(setter):
            setter(value)

    # ########## ACTION HANDLERS ########## #
    @dbus_propagate()
    def on_refresh_window(self, _, state_updates: Dict[str, Any], force: bool = False):
        if settings := state_updates.get("__settings__"):
            for k, v in settings.items():
                setattr(self.app_config, k, v)
            if (offline_mode := settings.get("offline_mode")) is not None:
                AdapterManager.on_offline_mode_change(offline_mode)

            del state_updates["__settings__"]
            self.app_config.save()

        if player_setting := state_updates.get("__player_setting__"):
            player_name, option_name, value = player_setting
            self.app_config.player_config[player_name][option_name] = value
            del state_updates["__player_setting__"]
            if pm := self.player_manager:
                pm.change_settings(self.app_config.player_config)
            self.app_config.save()

        for k, v in state_updates.items():
            setattr(self.app_config.state, k, v)
        self.update_window(force=force)

    def on_notification_closed(self, _):
        self.app_config.state.current_notification = None
        self.update_window()

    def on_add_new_music_provider(self, *args):
        self.show_configure_servers_dialog()

    def on_edit_current_music_provider(self, *args):
        self.show_configure_servers_dialog(self.app_config.provider.clone())

    def on_switch_music_provider(self, _, provider_id: GLib.Variant):
        if self.app_config.state.playing:
            self.on_play_pause()
        self.app_config.save()
        self.app_config.current_provider_id = provider_id.get_string()
        self.reset_state()
        self.app_config.save()

    def on_remove_music_provider(self, _, provider_id: GLib.Variant):
        provider = self.app_config.providers[provider_id.get_string()]
        confirm_dialog = Gtk.MessageDialog(
            transient_for=self.window,
            message_type=Gtk.MessageType.WARNING,
            buttons=(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_DELETE,
                Gtk.ResponseType.YES,
            ),
            text=f"Are you sure you want to delete the {provider.name} music provider?",
        )
        confirm_dialog.format_secondary_markup(
            "Deleting this music provider will delete all cached songs and metadata "
            "associated with this provider."
        )
        if confirm_dialog.run() == Gtk.ResponseType.YES:
            assert self.app_config.cache_location
            provider_dir = self.app_config.cache_location.joinpath(provider.id)
            shutil.rmtree(str(provider_dir), ignore_errors=True)
            del self.app_config.providers[provider.id]

        confirm_dialog.destroy()

    def on_window_go_to(self, win: Any, action: str, value: str):
        {
            "album": self.on_go_to_album,
            "artist": self.on_go_to_artist,
            "playlist": self.on_go_to_playlist,
        }[action](None, GLib.Variant("s", value))

    @dbus_propagate()
    def on_play_pause(self, *args):
        if self.app_config.state.current_song_index < 0:
            return

        self.app_config.state.playing = not self.app_config.state.playing

        if self.player_manager.song_loaded and (
            self.player_manager.current_song == self.app_config.state.current_song
        ):
            self.player_manager.toggle_play()
            self.save_play_queue()
        else:
            # This is from a restart, start playing the file.
            self.play_song(self.app_config.state.current_song_index)

        self.update_window()

    def on_next_track(self, *args):
        if self.app_config.state.current_song is None:
            # This may happen due to DBUS, ignore.
            return
        # Handle song repeating
        if self.app_config.state.repeat_type == RepeatType.REPEAT_SONG:
            song_index_to_play = self.app_config.state.current_song_index
        # Wrap around the play queue if at the end.
        elif (
            self.app_config.state.current_song_index
            == len(self.app_config.state.play_queue) - 1
        ):
            # This may happen due to D-Bus.
            if self.app_config.state.repeat_type == RepeatType.NO_REPEAT:
                return
            song_index_to_play = 0
        else:
            song_index_to_play = self.app_config.state.current_song_index + 1

        self.app_config.state.current_song_index = song_index_to_play
        self.app_config.state.song_progress = timedelta(0)
        if self.app_config.state.playing:
            self.play_song(song_index_to_play, reset=True)
        else:
            self.update_window()

    def on_prev_track(self, *args):
        if self.app_config.state.current_song is None:
            # This may happen due to DBUS, ignore.
            return
        # Go back to the beginning of the song if we are past 5 seconds.
        # Otherwise, go to the previous song.
        no_repeat = self.app_config.state.repeat_type == RepeatType.NO_REPEAT
        if self.app_config.state.repeat_type == RepeatType.REPEAT_SONG:
            song_index_to_play = self.app_config.state.current_song_index
        elif self.app_config.state.song_progress.total_seconds() < 5:
            if self.app_config.state.current_song_index == 0 and no_repeat:
                song_index_to_play = 0
            else:
                song_index_to_play = (
                    self.app_config.state.current_song_index - 1
                ) % len(self.app_config.state.play_queue)
        else:
            # Go back to the beginning of the song.
            song_index_to_play = self.app_config.state.current_song_index

        self.app_config.state.current_song_index = song_index_to_play
        self.app_config.state.song_progress = timedelta(0)
        if self.player_manager.playing:
            self.play_song(
                song_index_to_play,
                reset=True,
                # search backwards for a song to play if offline
                playable_song_search_direction=-1,
            )
        else:
            self.update_window()

    @dbus_propagate()
    def on_repeat_press(self, *args):
        # Cycle through the repeat types.
        new_repeat_type = RepeatType((self.app_config.state.repeat_type.value + 1) % 3)
        self.app_config.state.repeat_type = new_repeat_type
        self.update_window()

    @dbus_propagate()
    def on_shuffle_press(self, *args):
        if self.app_config.state.shuffle_on:
            # Revert to the old play queue.
            old_play_queue_copy = self.app_config.state.old_play_queue
            self.app_config.state.current_song_index = old_play_queue_copy.index(
                self.app_config.state.current_song.id
            )
            self.app_config.state.play_queue = old_play_queue_copy
        else:
            self.app_config.state.old_play_queue = self.app_config.state.play_queue

            mutable_play_queue = list(self.app_config.state.play_queue)

            # Remove the current song, then shuffle and put the song back.
            song_id = self.app_config.state.current_song.id
            del mutable_play_queue[self.app_config.state.current_song_index]
            random.shuffle(mutable_play_queue)
            self.app_config.state.play_queue = (song_id,) + tuple(mutable_play_queue)
            self.app_config.state.current_song_index = 0

        self.app_config.state.shuffle_on = not self.app_config.state.shuffle_on
        self.update_window()

    @dbus_propagate()
    def on_play_next(self, action: Any, song_ids: GLib.Variant):
        song_ids = tuple(song_ids)
        if self.app_config.state.current_song is None:
            insert_at = 0
        else:
            insert_at = self.app_config.state.current_song_index + 1

        self.app_config.state.play_queue = (
            self.app_config.state.play_queue[:insert_at]
            + song_ids
            + self.app_config.state.play_queue[insert_at:]
        )
        self.app_config.state.old_play_queue += song_ids
        self.update_window()

    @dbus_propagate()
    def on_add_to_queue(self, action: Any, song_ids: GLib.Variant):
        song_ids = tuple(song_ids)
        self.app_config.state.play_queue += tuple(song_ids)
        self.app_config.state.old_play_queue += tuple(song_ids)
        self.update_window()

    def on_go_to_album(self, action: Any, album_id: GLib.Variant):
        # Switch to the Alphabetical by Name view to guarantee that the album is there.
        self.app_config.state.current_album_search_query = AlbumSearchQuery(
            AlbumSearchQuery.Type.ALPHABETICAL_BY_NAME,
            genre=self.app_config.state.current_album_search_query.genre,
            year_range=self.app_config.state.current_album_search_query.year_range,
        )

        self.app_config.state.current_tab = "albums"
        self.app_config.state.selected_album_id = album_id.get_string()
        self.update_window()

    def on_go_to_artist(self, action: Any, artist_id: GLib.Variant):
        self.app_config.state.current_tab = "artists"
        self.app_config.state.selected_artist_id = artist_id.get_string()
        self.update_window()

    def browse_to(self, action: Any, item_id: GLib.Variant):
        self.app_config.state.current_tab = "browse"
        self.app_config.state.selected_browse_element_id = item_id.get_string()
        self.update_window()

    def on_go_to_playlist(self, action: Any, playlist_id: GLib.Variant):
        self.app_config.state.current_tab = "playlists"
        self.app_config.state.selected_playlist_id = playlist_id.get_string()
        self.update_window()

    def on_go_online(self, *args):
        self.on_refresh_window(None, {"__settings__": {"offline_mode": False}})

    def on_refresh_devices(self, *args):
        self.player_manager.refresh_players()

    def reset_state(self):
        if self.app_config.state.playing:
            self.on_play_pause()
        self.loading_state = True
        self.player_manager.reset()
        AdapterManager.reset(self.app_config, self.on_song_download_progress)
        self.loading_state = False

        # Update the window according to the new server configuration.
        self.update_window()

    def on_stack_change(self, stack: Gtk.Stack, _):
        self.app_config.state.current_tab = stack.get_visible_child_name()
        self.update_window()

    def on_song_clicked(
        self,
        win: Any,
        song_index: int,
        song_queue: Tuple[str, ...],
        metadata: Dict[str, Any],
    ):
        song_queue = tuple(song_queue)
        # Reset the play queue so that we don't ever revert back to the
        # previous one.
        old_play_queue = song_queue

        if (force_shuffle := metadata.get("force_shuffle_state")) is not None:
            self.app_config.state.shuffle_on = force_shuffle

        self.app_config.state.active_playlist_id = metadata.get("active_playlist_id")

        # If shuffle is enabled, then shuffle the playlist.
        if self.app_config.state.shuffle_on and not metadata.get("no_reshuffle"):
            song_id = song_queue[song_index]
            song_queue_list = list(
                song_queue[:song_index] + song_queue[song_index + 1 :]
            )
            random.shuffle(song_queue_list)
            song_queue = (song_id, *song_queue_list)
            song_index = 0

        self.play_song(
            song_index,
            reset=True,
            old_play_queue=old_play_queue,
            play_queue=song_queue,
        )

    def on_songs_removed(self, win: Any, song_indexes_to_remove: List[int]):
        self.app_config.state.play_queue = tuple(
            song_id
            for i, song_id in enumerate(self.app_config.state.play_queue)
            if i not in song_indexes_to_remove
        )

        # Determine how many songs before the currently playing one were also
        # deleted.
        before_current = [
            i
            for i in song_indexes_to_remove
            if i < self.app_config.state.current_song_index
        ]

        if self.app_config.state.current_song_index in song_indexes_to_remove:
            if len(self.app_config.state.play_queue) == 0:
                self.on_play_pause()
                self.app_config.state.current_song_index = -1
                self.update_window()
                return

            self.app_config.state.current_song_index -= len(before_current)
            self.play_song(self.app_config.state.current_song_index, reset=True)
        else:
            self.app_config.state.current_song_index -= len(before_current)
            self.update_window()
            self.save_play_queue()

    @dbus_propagate()
    def on_song_scrub(self, _, scrub_value: float):
        if not self.app_config.state.current_song or not self.window:
            return

        # This should not ever happen. The current_song should always have
        # a duration, but the Child object has `duration` optional because
        # it could be a directory.
        assert self.app_config.state.current_song.duration is not None
        new_time = self.app_config.state.current_song.duration * (scrub_value / 100)

        self.app_config.state.song_progress = new_time
        self.window.player_controls.update_scrubber(
            self.app_config.state.song_progress,
            self.app_config.state.current_song.duration,
            self.app_config.state.song_stream_cache_progress,
        )

        # If already playing, then make the player itself seek.
        if self.player_manager and self.player_manager.song_loaded:
            self.player_manager.seek(new_time)

        self.save_play_queue()

    def on_device_update(self, _, device_id: str):
        assert self.player_manager
        if device_id == self.app_config.state.current_device:
            return
        self.app_config.state.current_device = device_id

        if was_playing := self.app_config.state.playing:
            self.on_play_pause()

        self.player_manager.set_current_device_id(self.app_config.state.current_device)

        if self.dbus_manager:
            self.dbus_manager.property_diff()

        self.update_window()

        if was_playing:
            self.on_play_pause()
            if self.dbus_manager:
                self.dbus_manager.property_diff()

    @dbus_propagate()
    def on_mute_toggle(self, *args):
        self.app_config.state.is_muted = not self.app_config.state.is_muted
        self.player_manager.set_muted(self.app_config.state.is_muted)
        self.update_window()

    @dbus_propagate()
    def on_volume_change(self, _, value: float):
        assert self.player_manager
        self.app_config.state.volume = value
        self.player_manager.set_volume(self.app_config.state.volume)
        self.update_window()

    def on_window_key_press(self, window: Gtk.Window, event: Gdk.EventKey) -> bool:
        # Need to use bitwise & here to see if CTRL is pressed.
        if event.keyval == 102 and event.state & Gdk.ModifierType.CONTROL_MASK:
            # Ctrl + F
            window.search_entry.grab_focus()
            return False

        # Allow spaces to work in the text entry boxes.
        if (
            window.search_entry.has_focus()
            or window.playlists_panel.playlist_list.new_playlist_entry.has_focus()
        ):
            return False

        # Spacebar, home/prev
        keymap = {
            32: self.on_play_pause,
            65360: self.on_prev_track,
            65367: self.on_next_track,
        }

        action = keymap.get(event.keyval)
        if action:
            action()
            return True

        return False

    def on_song_download_progress(self, song_id: str, progress: DownloadProgress):
        assert self.window
        GLib.idle_add(self.window.update_song_download_progress, song_id, progress)

    def on_app_shutdown(self, app: "SublimeMusicApp"):
        self.exiting = True
        if glib_notify_exists:
            Notify.uninit()

        if tap_imported and self.tap:
            self.tap.stop()

        if self.app_config.provider is None:
            return

        if self.player_manager:
            if self.app_config.state.playing:
                self.save_play_queue()
            self.player_manager.pause()
            self.player_manager.shutdown()

        self.app_config.save()
        if self.dbus_manager:
            self.dbus_manager.shutdown()
        AdapterManager.shutdown()

    # ########## HELPER METHODS ########## #
    def show_configure_servers_dialog(
        self,
        provider_config: Optional[ProviderConfiguration] = None,
    ):
        """Show the Connect to Server dialog."""
        dialog = ConfigureProviderDialog(self.window, provider_config)
        result = dialog.run()
        if result == Gtk.ResponseType.APPLY:
            assert dialog.provider_config is not None
            provider_id = dialog.provider_config.id
            dialog.provider_config.persist_secrets()
            self.app_config.providers[provider_id] = dialog.provider_config
            self.app_config.save()

            if provider_id == self.app_config.current_provider_id:
                # Just update the window.
                self.update_window()
            else:
                # Switch to the new provider.
                if self.app_config.state.playing:
                    self.on_play_pause()
                self.app_config.current_provider_id = provider_id
                self.app_config.save()
                self.update_window(force=True)

        dialog.destroy()

    def update_window(self, force: bool = False):
        if not self.window:
            return
        logging.info(f"Updating window force={force}")
        GLib.idle_add(
            lambda: self.window.update(
                self.app_config, self.player_manager, force=force
            )
        )

    def update_play_state_from_server(self, prompt_confirm: bool = False):
        # TODO (#129): need to make the play queue list loading for the duration here if
        # prompt_confirm is False.
        if not prompt_confirm and self.app_config.state.playing:
            assert self.player_manager
            self.player_manager.pause()
            self.app_config.state.playing = False
            self.app_config.state.loading_play_queue = True
            self.update_window()

        def do_update(f: Result[PlayQueue]):
            play_queue = f.result()
            if not play_queue:
                self.app_config.state.loading_play_queue = False
                return
            play_queue.position = play_queue.position or timedelta(0)

            new_play_queue = tuple(s.id for s in play_queue.songs)
            new_song_progress = play_queue.position

            def do_resume(clear_notification: bool):
                assert self.player_manager
                if was_playing := self.app_config.state.playing:
                    self.on_play_pause()

                self.app_config.state.play_queue = new_play_queue
                self.app_config.state.song_progress = play_queue.position
                self.app_config.state.current_song_index = play_queue.current_index or 0
                self.app_config.state.loading_play_queue = False
                self.player_manager.reset()
                if clear_notification:
                    self.app_config.state.current_notification = None
                self.update_window()

                if was_playing:
                    self.on_play_pause()

            if prompt_confirm:
                # If there's not a significant enough difference in the song state,
                # don't prompt.
                progress_diff = 15.0
                if self.app_config.state.song_progress:
                    progress_diff = abs(
                        (
                            self.app_config.state.song_progress - new_song_progress
                        ).total_seconds()
                    )

                if (
                    self.app_config.state.play_queue == new_play_queue
                    and self.app_config.state.current_song
                ):
                    song_index = self.app_config.state.current_song_index
                    if song_index == play_queue.current_index and progress_diff < 15:
                        return

                # Show a notification to resume the play queue.
                resume_text = "Do you want to resume the play queue"
                if play_queue.changed_by or play_queue.changed:
                    resume_text += " saved"
                    if play_queue.changed_by:
                        resume_text += f" by {play_queue.changed_by}"
                    if play_queue.changed:
                        changed_str = play_queue.changed.astimezone(tz=None).strftime(
                            "%H:%M on %Y-%m-%d"
                        )
                        resume_text += f" at {changed_str}"
                resume_text += "?"

                self.app_config.state.current_notification = UIState.UINotification(
                    markup=f"<b>{resume_text}</b>",
                    actions=(("Resume", partial(do_resume, True)),),
                )
                self.update_window()

            else:  # just resume the play queue immediately
                do_resume(False)

        play_queue_future = AdapterManager.get_play_queue()
        play_queue_future.add_done_callback(lambda f: GLib.idle_add(do_update, f))

    song_playing_order_token = 0
    batch_download_jobs: Set[Result] = set()

    def play_song(
        self,
        song_index: int,
        reset: bool = False,
        old_play_queue: Tuple[str, ...] = None,
        play_queue: Tuple[str, ...] = None,
        playable_song_search_direction: int = 1,
    ):
        def do_reset():
            self.player_manager.reset()
            self.app_config.state.song_progress = timedelta(0)
            self.should_scrobble_song = True

        # Do this the old fashioned way so that we can have access to ``reset``
        # in the callback.
        @dbus_propagate(self)
        def do_play_song(order_token: int, song: Song):
            assert self.player_manager
            if order_token != self.song_playing_order_token:
                return

            uri = None
            try:
                if "file" in self.player_manager.supported_schemes:
                    uri = AdapterManager.get_song_file_uri(song)
            except CacheMissError:
                logging.debug("Couldn't find the file, will attempt to stream.")

            if not uri:
                try:
                    uri = AdapterManager.get_song_stream_uri(song)
                except Exception:
                    pass
                if (
                    not uri
                    or urlparse(uri).scheme not in self.player_manager.supported_schemes
                ):
                    self.app_config.state.current_notification = UIState.UINotification(
                        markup=f"<b>Unable to play {song.title}.</b>",
                        icon="dialog-error",
                    )
                    return

            # Prevent it from doing the thing where it continually loads
            # songs when it has to download.
            if reset:
                do_reset()

            # Start playing the song.
            if order_token != self.song_playing_order_token:
                return

            self.player_manager.play_media(
                uri,
                timedelta(0) if reset else self.app_config.state.song_progress,
                song,
            )
            self.app_config.state.playing = True
            self.update_window()

            # Show a song play notification.
            if self.app_config.song_play_notification:
                try:
                    if glib_notify_exists:
                        notification_lines = []
                        if album := song.album:
                            notification_lines.append(
                                f"<i>{bleach.clean(album.name)}</i>"
                            )
                        if artist := song.artist:
                            notification_lines.append(bleach.clean(artist.name))
                        song_notification = Notify.Notification.new(
                            song.title,
                            "\n".join(notification_lines),
                        )
                        song_notification.add_action(
                            "clicked",
                            "Open Sublime Music",
                            lambda *a: self.window.present() if self.window else None,
                        )
                        song_notification.show()

                        def on_cover_art_download_complete(cover_art_filename: str):
                            if order_token != self.song_playing_order_token:
                                return

                            # Add the image to the notification, and re-show
                            # the notification.
                            song_notification.set_image_from_pixbuf(
                                GdkPixbuf.Pixbuf.new_from_file_at_scale(
                                    cover_art_filename, 70, 70, True
                                )
                            )
                            song_notification.show()

                        cover_art_result = AdapterManager.get_cover_art_uri(
                            song.cover_art, "file"
                        )
                        cover_art_result.add_done_callback(
                            lambda f: on_cover_art_download_complete(f.result())
                        )

                    if sys.platform == "darwin":
                        notification_lines = []
                        if album := song.album:
                            notification_lines.append(album.name)
                        if artist := song.artist:
                            notification_lines.append(artist.name)
                        notification_text = "\n".join(notification_lines)
                        osascript_command = [
                            "display",
                            "notification",
                            f'"{notification_text}"',
                            "with",
                            "title",
                            f'"{song.title}"',
                        ]

                        os.system(f"osascript -e '{' '.join(osascript_command)}'")
                except Exception:
                    logging.warning(
                        "Unable to display notification. Is a notification daemon running?"  # noqa: E501
                    )

            # Download current song and prefetch songs. Only do this if the adapter can
            # download songs and allow_song_downloads is True and download_on_stream is
            # True.
            def on_song_download_complete(song_id: str):
                if order_token != self.song_playing_order_token:
                    return

                # Hotswap to the downloaded song.
                if (
                    # TODO (#182) allow hotswap if not playing. This requires being able
                    # to replace the currently playing URI with something different.
                    self.app_config.state.playing
                    and self.app_config.state.current_song
                    and self.app_config.state.current_song.id == song_id
                ):
                    # Switch to the local media if the player can hotswap without lag.
                    # For example, MPV can is barely noticable whereas there's quite a
                    # delay with Chromecast.
                    assert self.player_manager
                    if self.player_manager.can_start_playing_with_no_latency:
                        self.player_manager.play_media(
                            AdapterManager.get_song_file_uri(song),
                            self.app_config.state.song_progress,
                            song,
                        )

                # Always update the window
                self.update_window()

            if (
                # This only makes sense if the adapter is networked.
                AdapterManager.ground_truth_adapter_is_networked()
                # Don't download in offline mode.
                and not self.app_config.offline_mode
                and self.app_config.allow_song_downloads
                and self.app_config.download_on_stream
                and AdapterManager.can_batch_download_songs()
            ):
                song_ids = [song.id]

                # Add the prefetch songs.
                if (
                    repeat_type := self.app_config.state.repeat_type
                ) != RepeatType.REPEAT_SONG:
                    song_idx = self.app_config.state.play_queue.index(song.id)
                    is_repeat_queue = RepeatType.REPEAT_QUEUE == repeat_type
                    prefetch_idxs = []
                    for i in range(self.app_config.prefetch_amount):
                        prefetch_idx: int = song_idx + 1 + i
                        play_queue_len: int = len(self.app_config.state.play_queue)
                        if is_repeat_queue or prefetch_idx < play_queue_len:
                            prefetch_idxs.append(
                                prefetch_idx % play_queue_len  # noqa: S001
                            )
                    song_ids.extend(
                        [self.app_config.state.play_queue[i] for i in prefetch_idxs]
                    )

                self.batch_download_jobs.add(
                    AdapterManager.batch_download_songs(
                        song_ids,
                        before_download=lambda _: self.update_window(),
                        on_song_download_complete=on_song_download_complete,
                        one_at_a_time=True,
                        delay=5,
                    )
                )

        if old_play_queue:
            self.app_config.state.old_play_queue = old_play_queue

        if play_queue:
            self.app_config.state.play_queue = play_queue

        self.app_config.state.current_song_index = song_index

        for job in self.batch_download_jobs:
            job.cancel()

        self.song_playing_order_token += 1

        if play_queue:
            GLib.timeout_add(
                5000,
                partial(
                    self.save_play_queue,
                    song_playing_order_token=self.song_playing_order_token,
                ),
            )

        # If in offline mode, go to the first song in the play queue after the given
        # song that is actually playable.
        if self.app_config.offline_mode:
            statuses = AdapterManager.get_cached_statuses(
                self.app_config.state.play_queue
            )
            playable_statuses = (
                SongCacheStatus.CACHED,
                SongCacheStatus.PERMANENTLY_CACHED,
            )
            can_play = False
            current_song_index = self.app_config.state.current_song_index

            if statuses[current_song_index] in playable_statuses:
                can_play = True
            elif self.app_config.state.repeat_type != RepeatType.REPEAT_SONG:
                # See if any other songs in the queue are playable.
                play_queue_len = len(self.app_config.state.play_queue)
                cursor = (
                    current_song_index + playable_song_search_direction
                ) % play_queue_len
                for _ in range(play_queue_len):  # Don't infinite loop.
                    if self.app_config.state.repeat_type == RepeatType.NO_REPEAT:
                        if (
                            playable_song_search_direction == 1
                            and cursor < current_song_index
                        ) or (
                            playable_song_search_direction == -1
                            and cursor > current_song_index
                        ):
                            # We wrapped around to the end of the play queue without
                            # finding a song that can be played, and we aren't allowed
                            # to loop back.
                            break

                    # If we find a playable song, stop and play it.
                    if statuses[cursor] in playable_statuses:
                        self.play_song(cursor, reset)
                        return

                    cursor = (cursor + playable_song_search_direction) % play_queue_len

            if not can_play:
                # There are no songs that can be played. Show a notification that you
                # have to go online to play anything and then don't go further.
                if was_playing := self.app_config.state.playing:
                    self.on_play_pause()

                def go_online_clicked():
                    self.app_config.state.current_notification = None
                    self.on_go_online()
                    if was_playing:
                        self.on_play_pause()

                if all(s == SongCacheStatus.NOT_CACHED for s in statuses):
                    markup = (
                        "<b>None of the songs in your play queue are cached for "
                        "offline playback.</b>\nGo online to start playing your queue."
                    )
                else:
                    markup = (
                        "<b>None of the remaining songs in your play queue are cached "
                        "for offline playback.</b>\nGo online to contiue playing your "
                        "queue."
                    )

                self.app_config.state.current_notification = UIState.UINotification(
                    icon="cloud-offline-symbolic",
                    markup=markup,
                    actions=(("Go Online", go_online_clicked),),
                )
                if reset:
                    do_reset()
                self.update_window()
                return

        song_details_future = AdapterManager.get_song_details(
            self.app_config.state.play_queue[self.app_config.state.current_song_index]
        )
        if song_details_future.data_is_available:
            song_details_future.add_done_callback(
                lambda f: do_play_song(self.song_playing_order_token, f.result())
            )
        else:
            song_details_future.add_done_callback(
                lambda f: GLib.idle_add(
                    partial(do_play_song, self.song_playing_order_token), f.result()
                ),
            )

    def save_play_queue(self, song_playing_order_token: int = None):
        if (
            len(self.app_config.state.play_queue) == 0
            or self.app_config.provider is None
            or (
                song_playing_order_token
                and song_playing_order_token != self.song_playing_order_token
            )
        ):
            return

        position = self.app_config.state.song_progress
        self.last_play_queue_update = position or timedelta(0)

        if AdapterManager.can_save_play_queue() and self.app_config.state.current_song:
            AdapterManager.save_play_queue(
                song_ids=self.app_config.state.play_queue,
                current_song_index=self.app_config.state.current_song_index,
                position=position,
            )



import threading
from datetime import timedelta
from typing import Callable, cast, Dict, Optional, Tuple, Type, Union

import mpv

from .base import Player, PlayerDeviceEvent, PlayerEvent
from ..adapters.api_objects import Song

REPLAY_GAIN_KEY = "Replay Gain"


class MPVPlayer(Player):
    enabled = True
    name = "Local Playback"
    can_start_playing_with_no_latency = True
    supported_schemes = {"http", "https", "file"}
    song_loaded = False

    _progress_value_lock = threading.Lock()
    _progress_value_count = 0

    _volume = 100.0
    _muted = False

    _is_mock = False

    @staticmethod
    def get_configuration_options() -> Dict[str, Union[Type, Tuple[str, ...]]]:
        return {REPLAY_GAIN_KEY: ("Disabled", "Track", "Album")}

    def __init__(
        self,
        on_timepos_change: Callable[[Optional[float]], None],
        on_track_end: Callable[[], None],
        on_player_event: Callable[[PlayerEvent], None],
        player_device_change_callback: Callable[[PlayerDeviceEvent], None],
        config: Dict[str, Union[str, int, bool]],
    ):
        self.mpv = mpv.MPV()
        if MPVPlayer._is_mock:
            self.mpv.audio_device = "null"
        self.mpv.audio_client_name = "sublime-music"
        self.change_settings(config)

        @self.mpv.property_observer("time-pos")
        def time_observer(_, value: Optional[float]):
            on_timepos_change(value)
            if value is None and self._progress_value_count > 1:
                on_track_end()
                with self._progress_value_lock:
                    self._progress_value_count = 0

            if value:
                with self._progress_value_lock:
                    self._progress_value_count += 1

        @self.mpv.property_observer("demuxer-cache-time")
        def cache_size_observer(_, value: Optional[float]):
            on_player_event(
                PlayerEvent(
                    PlayerEvent.EventType.STREAM_CACHE_PROGRESS_CHANGE,
                    "this device",
                    stream_cache_duration=value,
                )
            )

        # Indicate to the UI that we exist.
        player_device_change_callback(
            PlayerDeviceEvent(
                PlayerDeviceEvent.Delta.ADD, type(self), "this device", "This Device"
            )
        )

    def change_settings(self, config: Dict[str, Union[str, int, bool]]):
        self.config = config
        self.mpv.replaygain = {
            "Disabled": "no",
            "Track": "track",
            "Album": "album",
        }.get(cast(str, config.get(REPLAY_GAIN_KEY, "Disabled")), "no")

    def refresh_players(self):
        # Don't do anything
        pass

    def set_current_device_id(self, device_id: str):
        # Don't do anything beacuse it should always be the "this device" ID.
        pass

    def shutdown(self):
        pass

    def reset(self):
        self.song_loaded = False
        with self._progress_value_lock:
            self._progress_value_count = 0

    @property
    def playing(self) -> bool:
        return not self.mpv.pause

    def get_volume(self) -> float:
        return self._volume

    def set_volume(self, volume: float):
        if not self._muted:
            self.mpv.volume = volume
        self._volume = volume

    def get_is_muted(self) -> bool:
        return self._muted

    def set_muted(self, muted: bool):
        self.mpv.volume = 0 if muted else self._volume
        self._muted = muted

    def play_media(self, uri: str, progress: timedelta, song: Song):
        with self._progress_value_lock:
            self._progress_value_count = 0

        options = {
            "force-seekable": "yes",
            "start": str(progress.total_seconds()),
        }
        self.mpv.command(
            "loadfile", uri, "replace", ",".join(f"{k}={v}" for k, v in options.items())
        )
        self.mpv.pause = False
        self.song_loaded = True

    def pause(self):
        self.mpv.pause = True

    def play(self):
        self.mpv.pause = False

    def seek(self, position: timedelta):
        self.mpv.seek(str(position.total_seconds()), "absolute")



from dataclasses import dataclass, field
from datetime import timedelta
from enum import Enum
from typing import Any, Callable, Dict, Optional, Set, Tuple, Type

from ..adapters import AlbumSearchQuery
from ..adapters.api_objects import Genre, Song


class RepeatType(Enum):
    NO_REPEAT = 0
    REPEAT_QUEUE = 1
    REPEAT_SONG = 2

    @property
    def icon(self) -> str:
        """
        Get the icon for the repeat type.

        >>> RepeatType.NO_REPEAT.icon, RepeatType.REPEAT_QUEUE.icon
        ('media-playlist-repeat-symbolic', 'media-playlist-repeat-symbolic')
        >>> RepeatType.REPEAT_SONG.icon
        'media-playlist-repeat-song-symbolic'
        """
        song_str = "-song" if self == RepeatType.REPEAT_SONG else ""
        return f"media-playlist-repeat{song_str}-symbolic"

    def as_mpris_loop_status(self) -> str:
        return ["None", "Playlist", "Track"][self.value]

    @staticmethod
    def from_mpris_loop_status(loop_status: str) -> "RepeatType":
        return {
            "None": RepeatType.NO_REPEAT,
            "Track": RepeatType.REPEAT_SONG,
            "Playlist": RepeatType.REPEAT_QUEUE,
        }[loop_status]


@dataclass
class UIState:
    """Represents the UI state of the application."""

    @dataclass(unsafe_hash=True)
    class UINotification:
        markup: str
        actions: Tuple[Tuple[str, Callable[[], None]], ...] = field(
            default_factory=tuple
        )
        icon: Optional[str] = None

    version: int = 1

    # Play state
    playing: bool = False
    current_song_index: int = -1
    play_queue: Tuple[str, ...] = field(default_factory=tuple)
    old_play_queue: Tuple[str, ...] = field(default_factory=tuple)
    _volume: Dict[str, float] = field(default_factory=lambda: {"this device": 100.0})
    is_muted: bool = False
    repeat_type: RepeatType = RepeatType.NO_REPEAT
    shuffle_on: bool = False
    song_progress: timedelta = timedelta()
    song_stream_cache_progress: Optional[timedelta] = timedelta()
    current_device: str = "this device"
    connecting_to_device: bool = False
    connected_device_name: Optional[str] = None
    available_players: Dict[Type, Set[Tuple[str, str]]] = field(default_factory=dict)

    # UI state
    current_tab: str = "albums"
    selected_album_id: Optional[str] = None
    selected_artist_id: Optional[str] = None
    selected_browse_element_id: Optional[str] = None
    selected_playlist_id: Optional[str] = None
    album_sort_direction: str = "ascending"
    album_page_size: int = 30
    album_page: int = 0
    current_notification: Optional[UINotification] = None
    playlist_details_expanded: bool = True
    artist_details_expanded: bool = True
    loading_play_queue: bool = False

    # State for Album sort.
    class _DefaultGenre(Genre):
        def __init__(self):
            self.name = "Rock"

    current_album_search_query: AlbumSearchQuery = AlbumSearchQuery(
        AlbumSearchQuery.Type.RANDOM,
        genre=_DefaultGenre(),
        year_range=(2010, 2020),
    )

    active_playlist_id: Optional[str] = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["song_stream_cache_progress"]
        del state["current_notification"]
        del state["playing"]
        del state["available_players"]
        return state

    def __setstate__(self, state: Dict[str, Any]):
        self.__dict__.update(state)
        self.song_stream_cache_progress = None
        self.current_notification = None
        self.playing = False

    def __init_available_players__(self):
        from sublime_music.players import PlayerManager

        self.available_players = {
            pt: set() for pt in PlayerManager.available_player_types
        }

    def migrate(self):
        pass

    _current_song: Optional[Song] = None

    @property
    def current_song(self) -> Optional[Song]:
        if not self.play_queue or self.current_song_index < 0:
            return None

        from sublime_music.adapters import AdapterManager

        current_song_id = self.play_queue[self.current_song_index]

        if not self._current_song or self._current_song.id != current_song_id:
            self._current_song = AdapterManager.get_song_details(
                current_song_id
            ).result()

        return self._current_song

    @property
    def volume(self) -> float:
        return self._volume.get(self.current_device, 100.0)

    @volume.setter
    def volume(self, value: float):
        self._volume[self.current_device] = value



import json
import shutil
from dataclasses import asdict
from datetime import timedelta
from pathlib import Path
from typing import Any, cast, Generator, Iterable, Tuple

import pytest

from peewee import SelectQuery

from sublime_music.adapters import (
    AlbumSearchQuery,
    api_objects as SublimeAPI,
    CacheMissError,
    SongCacheStatus,
)
from sublime_music.adapters.filesystem import FilesystemAdapter
from sublime_music.adapters.subsonic import api_objects as SubsonicAPI

MOCK_DATA_FILES = Path(__file__).parent.joinpath("mock_data")
MOCK_ALBUM_ART = MOCK_DATA_FILES.joinpath("album-art.png")
MOCK_ALBUM_ART2 = MOCK_DATA_FILES.joinpath("album-art2.png")
MOCK_ALBUM_ART3 = MOCK_DATA_FILES.joinpath("album-art3.png")
MOCK_SONG_FILE = MOCK_DATA_FILES.joinpath("test-song.mp3")
MOCK_SONG_FILE2 = MOCK_DATA_FILES.joinpath("test-song2.mp3")
MOCK_ALBUM_ART_HASH = "5d7bee4f3fe25b18cd2a66f1c9767e381bc64328"
MOCK_ALBUM_ART2_HASH = "031a8a1ca01f64f851a22d5478e693825a00fb23"
MOCK_ALBUM_ART3_HASH = "46a8af0f8fe370e59202a545803e8bbb3a4a41ee"
MOCK_SONG_FILE_HASH = "fe12d0712dbfd6ff7f75ef3783856a7122a78b0a"
MOCK_SONG_FILE2_HASH = "c32597c724e2e484dbf5856930b2e5bb80de13b7"

MOCK_SUBSONIC_SONGS = [
    SubsonicAPI.Song(
        "2",
        title="Song 2",
        parent_id="d1",
        _album="foo",
        album_id="a1",
        _artist="cool",
        artist_id="art1",
        duration=timedelta(seconds=20.8),
        path="foo/song2.mp3",
        cover_art="s2",
        _genre="Bar",
    ),
    SubsonicAPI.Song(
        "1",
        title="Song 1",
        parent_id="d1",
        _album="foo",
        album_id="a1",
        _artist="foo",
        artist_id="art2",
        duration=timedelta(seconds=10.2),
        path="foo/song1.mp3",
        cover_art="s1",
        _genre="Foo",
    ),
    SubsonicAPI.Song(
        "1",
        title="Song 1",
        parent_id="d1",
        _album="foo",
        album_id="a1",
        _artist="foo",
        artist_id="art2",
        duration=timedelta(seconds=10.2),
        path="foo/song1.mp3",
        cover_art="s1",
        _genre="Foo",
    ),
]

KEYS = FilesystemAdapter.CachedDataKey


@pytest.fixture
def adapter(tmp_path: Path):
    adapter = FilesystemAdapter({}, tmp_path)
    yield adapter
    adapter.shutdown()


@pytest.fixture
def cache_adapter(tmp_path: Path):
    adapter = FilesystemAdapter({}, tmp_path, is_cache=True)
    yield adapter
    adapter.shutdown()


def mock_data_files(
    request_name: str,
    mode: str = "r",
) -> Generator[Tuple[Path, Any], None, None]:
    """
    Yields all of the files in the mock_data directory that start with ``request_name``.
    """
    for file in MOCK_DATA_FILES.iterdir():
        if file.name.split("-")[0] in request_name:
            with open(file, mode) as f:
                yield file, f.read()


def verify_songs(
    actual_songs: Iterable[SublimeAPI.Song], expected_songs: Iterable[SubsonicAPI.Song]
):
    actual_songs, expected_songs = (list(actual_songs), list(expected_songs))
    assert len(actual_songs) == len(expected_songs)
    for actual, song in zip(actual_songs, expected_songs):
        for k, v in asdict(song).items():
            if k in ("_genre", "_album", "_artist", "album_id", "artist_id"):
                continue
            print(k, "->", v)  # noqa: T001

            actual_value = getattr(actual, k, None)

            if k == "album":
                assert ("a1", "foo") == (actual_value.id, actual_value.name)
            elif k == "genre":
                assert v["name"] == actual_value.name
            elif k == "artist":
                assert (v["id"], v["name"]) == (actual_value.id, actual_value.name)
            else:
                assert actual_value == v


def test_caching_get_playlists(cache_adapter: FilesystemAdapter):
    with pytest.raises(CacheMissError):
        cache_adapter.get_playlists()

    # Ingest an empty list (for example, no playlists added yet to server).
    cache_adapter.ingest_new_data(KEYS.PLAYLISTS, None, [])

    # After the first cache miss of get_playlists, even if an empty list is
    # returned, the next one should not be a cache miss.
    cache_adapter.get_playlists()

    # Ingest two playlists.
    cache_adapter.ingest_new_data(
        KEYS.PLAYLISTS,
        None,
        [
            SubsonicAPI.Playlist("1", "test1", comment="comment"),
            SubsonicAPI.Playlist("2", "test2"),
        ],
    )

    playlists = cache_adapter.get_playlists()
    assert len(playlists) == 2
    assert (playlists[0].id, playlists[0].name, playlists[0].comment) == (
        "1",
        "test1",
        "comment",
    )
    assert (playlists[1].id, playlists[1].name) == ("2", "test2")

    # Ingest a new playlist list with one of them deleted.
    cache_adapter.ingest_new_data(
        KEYS.PLAYLISTS,
        None,
        [
            SubsonicAPI.Playlist("1", "test1", comment="comment"),
            SubsonicAPI.Playlist("3", "test3"),
        ],
    )

    # Now, Playlist 2 should be gone.
    playlists = cache_adapter.get_playlists()
    assert len(playlists) == 2
    assert (playlists[0].id, playlists[0].name, playlists[0].comment) == (
        "1",
        "test1",
        "comment",
    )
    assert (playlists[1].id, playlists[1].name) == ("3", "test3")


def test_no_caching_get_playlists(adapter: FilesystemAdapter):
    adapter.get_playlists()

    # TODO (#188): Create a playlist (that should be allowed only if this is acting as
    # a ground truth adapter)
    # cache_adapter.create_playlist()

    adapter.get_playlists()
    # TODO (#188): verify playlist


def test_caching_get_playlist_details(cache_adapter: FilesystemAdapter):
    with pytest.raises(CacheMissError):
        cache_adapter.get_playlist_details("1")

    # Simulate the playlist being retrieved from Subsonic.
    cache_adapter.ingest_new_data(
        KEYS.PLAYLIST_DETAILS,
        "1",
        SubsonicAPI.Playlist("1", "test1", songs=MOCK_SUBSONIC_SONGS[:2]),
    )

    playlist = cache_adapter.get_playlist_details("1")
    assert playlist.id == "1"
    assert playlist.name == "test1"
    assert playlist.song_count == 2
    assert playlist.duration == timedelta(seconds=31)
    verify_songs(playlist.songs, MOCK_SUBSONIC_SONGS[:2])

    # "Force refresh" the playlist and add a new song (duplicate).
    cache_adapter.ingest_new_data(
        KEYS.PLAYLIST_DETAILS,
        "1",
        SubsonicAPI.Playlist("1", "foo", songs=MOCK_SUBSONIC_SONGS),
    )

    playlist = cache_adapter.get_playlist_details("1")
    assert playlist.id == "1"
    assert playlist.name == "foo"
    assert playlist.song_count == 3
    assert playlist.duration == timedelta(seconds=41.2)
    verify_songs(playlist.songs, MOCK_SUBSONIC_SONGS)

    with pytest.raises(CacheMissError):
        cache_adapter.get_playlist_details("2")

    # Now ingest the playlist list and make sure that it doesn't override the songs in
    # the first Playlist.
    cache_adapter.ingest_new_data(
        KEYS.PLAYLISTS,
        None,
        [
            SubsonicAPI.Playlist(
                "1", "foo", song_count=3, duration=timedelta(seconds=41.2)
            ),
            SubsonicAPI.Playlist(
                "3", "test3", song_count=3, duration=timedelta(seconds=30)
            ),
        ],
    )

    playlist = cache_adapter.get_playlist_details("1")
    verify_songs(playlist.songs, MOCK_SUBSONIC_SONGS)


def test_no_caching_get_playlist_details(adapter: FilesystemAdapter):
    with pytest.raises(Exception):
        adapter.get_playlist_details("1")

    # TODO (#188): Create a playlist (that should be allowed only if this is acting as
    # a ground truth adapter)
    # cache_adapter.create_playlist()

    # adapter.get_playlist_details('1')
    # TODO (#188): verify playlist details


def test_caching_get_playlist_then_details(cache_adapter: FilesystemAdapter):
    # Ingest a list of playlists (like the sidebar, without songs)
    cache_adapter.ingest_new_data(
        KEYS.PLAYLISTS,
        None,
        [SubsonicAPI.Playlist("1", "test1"), SubsonicAPI.Playlist("2", "test2")],
    )

    # Trying to get playlist details should generate a cache miss, but should
    # include the data that we know about.
    try:
        cache_adapter.get_playlist_details("1")
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data
        assert e.partial_data.id == "1"
        assert e.partial_data.name == "test1"

    # Simulate getting playlist details for id=1, then id=2
    cache_adapter.ingest_new_data(
        KEYS.PLAYLIST_DETAILS,
        "1",
        SubsonicAPI.Playlist("1", "test1"),
    )

    cache_adapter.ingest_new_data(
        KEYS.PLAYLIST_DETAILS,
        "2",
        SubsonicAPI.Playlist("2", "test2", songs=MOCK_SUBSONIC_SONGS),
    )

    # Going back and getting playlist details for the first one should not
    # cache miss.
    playlist = cache_adapter.get_playlist_details("1")
    assert playlist.id == "1"
    assert playlist.name == "test1"


def test_cache_cover_art(cache_adapter: FilesystemAdapter):
    with pytest.raises(CacheMissError):
        cache_adapter.get_cover_art_uri("pl_test1", "file", size=300)

    # After ingesting the data, reading from the cache should give the exact same file.
    cache_adapter.ingest_new_data(KEYS.COVER_ART_FILE, "pl_test1", MOCK_ALBUM_ART)
    with open(
        cache_adapter.get_cover_art_uri("pl_test1", "file", size=300), "wb+"
    ) as cached:
        with open(MOCK_ALBUM_ART, "wb+") as expected:
            assert cached.read() == expected.read()


def test_invalidate_playlist(cache_adapter: FilesystemAdapter):
    cache_adapter.ingest_new_data(
        KEYS.PLAYLISTS,
        None,
        [SubsonicAPI.Playlist("1", "test1"), SubsonicAPI.Playlist("2", "test2")],
    )
    cache_adapter.ingest_new_data(
        KEYS.COVER_ART_FILE,
        "pl_test1",
        MOCK_ALBUM_ART,
    )
    cache_adapter.ingest_new_data(
        KEYS.PLAYLIST_DETAILS,
        "2",
        SubsonicAPI.Playlist("2", "test2", cover_art="pl_2", songs=[]),
    )
    cache_adapter.ingest_new_data(
        KEYS.COVER_ART_FILE,
        "pl_2",
        MOCK_ALBUM_ART2,
    )

    stale_uri_1 = cache_adapter.get_cover_art_uri("pl_test1", "file", size=300)
    stale_uri_2 = cache_adapter.get_cover_art_uri("pl_2", "file", size=300)

    cache_adapter.invalidate_data(KEYS.PLAYLISTS, None)
    cache_adapter.invalidate_data(KEYS.PLAYLIST_DETAILS, "2")
    cache_adapter.invalidate_data(KEYS.COVER_ART_FILE, "pl_test1")

    # After invalidating the data, it should cache miss, but still have the old, stale,
    # data.
    try:
        cache_adapter.get_playlists()
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data
        assert len(e.partial_data) == 2

    try:
        cache_adapter.get_cover_art_uri("pl_test1", "file", size=300)
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data
        assert e.partial_data == stale_uri_1

    try:
        cache_adapter.get_playlist_details("2")
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data

    # Even though the pl_2 cover art file wasn't explicitly invalidated, it should have
    # been invalidated with the playlist details invalidation.
    try:
        cache_adapter.get_cover_art_uri("pl_2", "file", size=300)
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data
        assert e.partial_data == stale_uri_2


def test_invalidate_song_file(cache_adapter: FilesystemAdapter):
    cache_adapter.ingest_new_data(KEYS.SONG, "2", MOCK_SUBSONIC_SONGS[0])
    cache_adapter.ingest_new_data(KEYS.SONG, "1", MOCK_SUBSONIC_SONGS[1])
    cache_adapter.ingest_new_data(
        KEYS.COVER_ART_FILE,
        "s1",
        MOCK_ALBUM_ART,
    )
    cache_adapter.ingest_new_data(KEYS.SONG_FILE, "1", (None, MOCK_SONG_FILE, None))
    cache_adapter.ingest_new_data(KEYS.SONG_FILE, "2", (None, MOCK_SONG_FILE2, None))

    cache_adapter.invalidate_data(KEYS.SONG_FILE, "1")
    cache_adapter.invalidate_data(KEYS.COVER_ART_FILE, "s1")

    with pytest.raises(CacheMissError):
        cache_adapter.get_song_file_uri("1", "file")

    with pytest.raises(CacheMissError):
        cache_adapter.get_cover_art_uri("s1", "file", size=300)

    # Make sure it didn't delete the other song.
    assert cache_adapter.get_song_file_uri("2", "file").endswith("song2.mp3")


def test_malformed_song_path(cache_adapter: FilesystemAdapter):
    cache_adapter.ingest_new_data(KEYS.SONG, "1", MOCK_SUBSONIC_SONGS[1])
    cache_adapter.ingest_new_data(KEYS.SONG, "2", MOCK_SUBSONIC_SONGS[0])
    cache_adapter.ingest_new_data(
        KEYS.SONG_FILE, "1", ("/malformed/path", MOCK_SONG_FILE, None)
    )
    cache_adapter.ingest_new_data(
        KEYS.SONG_FILE, "2", ("fine/path/song2.mp3", MOCK_SONG_FILE2, None)
    )

    song_uri = cache_adapter.get_song_file_uri("1", "file")
    assert song_uri.endswith(f"/music/{MOCK_SONG_FILE_HASH}")

    song_uri2 = cache_adapter.get_song_file_uri("2", "file")
    assert song_uri2.endswith("fine/path/song2.mp3")


def test_get_cached_statuses(cache_adapter: FilesystemAdapter):
    cache_adapter.ingest_new_data(KEYS.SONG, "1", MOCK_SUBSONIC_SONGS[1])
    assert cache_adapter.get_cached_statuses(["1"]) == {"1": SongCacheStatus.NOT_CACHED}

    cache_adapter.ingest_new_data(KEYS.SONG_FILE, "1", (None, MOCK_SONG_FILE, None))
    assert cache_adapter.get_cached_statuses(["1"]) == {"1": SongCacheStatus.CACHED}

    cache_adapter.ingest_new_data(KEYS.SONG_FILE_PERMANENT, "1", None)
    assert cache_adapter.get_cached_statuses(["1"]) == {
        "1": SongCacheStatus.PERMANENTLY_CACHED
    }

    cache_adapter.invalidate_data(KEYS.SONG_FILE, "1")
    assert cache_adapter.get_cached_statuses(["1"]) == {
        "1": SongCacheStatus.CACHED_STALE
    }

    cache_adapter.delete_data(KEYS.SONG_FILE, "1")
    assert cache_adapter.get_cached_statuses(["1"]) == {"1": SongCacheStatus.NOT_CACHED}


def test_delete_playlists(cache_adapter: FilesystemAdapter):
    cache_adapter.ingest_new_data(
        KEYS.PLAYLIST_DETAILS,
        "1",
        SubsonicAPI.Playlist("1", "test1", cover_art="pl_1", songs=[]),
    )
    cache_adapter.ingest_new_data(
        KEYS.PLAYLIST_DETAILS,
        "2",
        SubsonicAPI.Playlist("2", "test1", cover_art="pl_2", songs=[]),
    )
    cache_adapter.ingest_new_data(
        KEYS.COVER_ART_FILE,
        "pl_1",
        MOCK_ALBUM_ART,
    )

    # Deleting a playlist should get rid of it entirely.
    cache_adapter.delete_data(KEYS.PLAYLIST_DETAILS, "2")
    try:
        cache_adapter.get_playlist_details("2")
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data is None

    # Deleting a playlist with associated cover art should get rid the cover art too.
    cache_adapter.delete_data(KEYS.PLAYLIST_DETAILS, "1")
    try:
        cache_adapter.get_cover_art_uri("pl_1", "file", size=300)
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data is None

    # Even if the cover art failed to be deleted, it should cache miss.
    shutil.copy(
        MOCK_ALBUM_ART,
        str(cache_adapter.cover_art_dir.joinpath(MOCK_ALBUM_ART_HASH)),
    )
    try:
        cache_adapter.get_cover_art_uri("pl_1", "file", size=300)
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data is None


def test_delete_song_data(cache_adapter: FilesystemAdapter):
    cache_adapter.ingest_new_data(KEYS.SONG, "1", MOCK_SUBSONIC_SONGS[1])
    cache_adapter.ingest_new_data(KEYS.SONG_FILE, "1", (None, MOCK_SONG_FILE, None))
    cache_adapter.ingest_new_data(
        KEYS.COVER_ART_FILE,
        "s1",
        MOCK_ALBUM_ART,
    )

    music_file_path = cache_adapter.get_song_file_uri("1", "file")
    cover_art_path = cache_adapter.get_cover_art_uri("s1", "file", size=300)

    cache_adapter.delete_data(KEYS.SONG_FILE, "1")
    cache_adapter.delete_data(KEYS.COVER_ART_FILE, "s1")

    assert not Path(music_file_path).exists()
    assert not Path(cover_art_path).exists()

    try:
        cache_adapter.get_song_file_uri("1", "file")
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data is None

    try:
        cache_adapter.get_cover_art_uri("s1", "file", size=300)
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data is None


def test_caching_get_genres(cache_adapter: FilesystemAdapter):
    with pytest.raises(CacheMissError):
        cache_adapter.get_genres()

    cache_adapter.ingest_new_data(KEYS.SONG, "2", MOCK_SUBSONIC_SONGS[0])
    cache_adapter.ingest_new_data(KEYS.SONG, "1", MOCK_SUBSONIC_SONGS[1])

    # Getting genres now should look at what's on the songs. This sould cache miss, but
    # still give some data.
    try:
        cache_adapter.get_genres()
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert [g.name for g in cast(Iterable, e.partial_data)] == ["Bar", "Foo"]

    # After we actually ingest the actual list, it should be returned instead.
    cache_adapter.ingest_new_data(
        KEYS.GENRES,
        None,
        [
            SubsonicAPI.Genre("Bar", 10, 20),
            SubsonicAPI.Genre("Baz", 10, 20),
            SubsonicAPI.Genre("Foo", 10, 20),
        ],
    )
    assert {g.name for g in cache_adapter.get_genres()} == {"Bar", "Baz", "Foo"}


def test_caching_get_song_details(cache_adapter: FilesystemAdapter):
    with pytest.raises(CacheMissError):
        cache_adapter.get_song_details("1")

    # Simulate the song details being retrieved from Subsonic.
    cache_adapter.ingest_new_data(KEYS.SONG, "1", MOCK_SUBSONIC_SONGS[1])

    song = cache_adapter.get_song_details("1")
    assert song.id == "1"
    assert song.title == "Song 1"
    assert song.album
    assert (song.album.id, song.album.name) == ("a1", "foo")
    assert song.artist and song.artist.name == "foo"
    assert song.parent_id == "d1"
    assert song.duration == timedelta(seconds=10.2)
    assert song.path == "foo/song1.mp3"
    assert song.genre and song.genre.name == "Foo"

    # "Force refresh" the song details
    cache_adapter.ingest_new_data(
        KEYS.SONG,
        "1",
        SubsonicAPI.Song(
            "1",
            title="Song 1",
            parent_id="bar",
            _album="bar",
            album_id="a2",
            _artist="bar",
            artist_id="art2",
            duration=timedelta(seconds=10.2),
            path="bar/song1.mp3",
            _genre="Bar",
        ),
    )

    song = cache_adapter.get_song_details("1")
    assert song.id == "1"
    assert song.title == "Song 1"
    assert song.album and song.artist
    assert (song.album.id, song.album.name) == ("a2", "bar")
    assert (song.artist.id, song.artist.name) == ("art2", "bar")
    assert song.parent_id == "bar"
    assert song.duration == timedelta(seconds=10.2)
    assert song.path == "bar/song1.mp3"
    assert song.genre and song.genre.name == "Bar"

    with pytest.raises(CacheMissError):
        cache_adapter.get_playlist_details("2")


def test_caching_get_song_details_missing_data(cache_adapter: FilesystemAdapter):
    with pytest.raises(CacheMissError):
        cache_adapter.get_song_details("1")

    # Ingest a song without an album ID and artist ID, but with album and artist name.
    cache_adapter.ingest_new_data(
        KEYS.SONG,
        "1",
        SubsonicAPI.Song(
            "1",
            title="Song 1",
            parent_id="bar",
            _album="bar",
            _artist="foo",
            duration=timedelta(seconds=10.2),
            path="foo/bar/song1.mp3",
            _genre="Bar",
        ),
    )

    song = cache_adapter.get_song_details("1")
    assert song.id == "1"
    assert song.title == "Song 1"
    assert song.album
    assert (song.album.id, song.album.name) == (
        "invalid:62cdb7020ff920e5aa642c3d4066950dd1f01f4d",
        "bar",
    )
    assert (song.artist.id, song.artist.name) == (
        "invalid:0beec7b5ea3f0fdbc95d0dd47f3c5bc275da8a33",
        "foo",
    )
    assert song.parent_id == "bar"
    assert song.duration == timedelta(seconds=10.2)
    assert song.path == "foo/bar/song1.mp3"
    assert song.genre and song.genre.name == "Bar"

    # Because the album and artist are invalid (doesn't have an album/artist ID), it
    # shouldn't show up in any results.
    try:
        list(
            cache_adapter.get_albums(
                AlbumSearchQuery(AlbumSearchQuery.Type.ALPHABETICAL_BY_NAME)
            )
        )
    except CacheMissError as e:
        assert e.partial_data is not None
        assert len(e.partial_data) == 0

    albums = list(cache_adapter.get_all_albums())
    assert len(albums) == 0

    with pytest.raises(CacheMissError):
        cache_adapter.get_album("invalid:62cdb7020ff920e5aa642c3d4066950dd1f01f4d")

    try:
        list(cache_adapter.get_artists())
    except CacheMissError as e:
        assert e.partial_data is not None
        assert len(e.partial_data) == 0

    with pytest.raises(CacheMissError):
        cache_adapter.get_artist("invalid:0beec7b5ea3f0fdbc95d0dd47f3c5bc275da8a33")


def test_caching_less_info(cache_adapter: FilesystemAdapter):
    cache_adapter.ingest_new_data(
        KEYS.SONG,
        "1",
        SubsonicAPI.Song(
            "1",
            title="Song 1",
            parent_id="bar",
            _album="bar",
            album_id="a2",
            _artist="bar",
            artist_id="art2",
            duration=timedelta(seconds=10.2),
            path="bar/song1.mp3",
            _genre="Bar",
        ),
    )
    cache_adapter.ingest_new_data(
        KEYS.SONG,
        "1",
        SubsonicAPI.Song(
            "1",
            title="Song 1",
            parent_id="bar",
            duration=timedelta(seconds=10.2),
            path="bar/song1.mp3",
        ),
    )

    song = cache_adapter.get_song_details("1")
    assert song.album and song.album.name == "bar"
    assert song.artist and song.artist.name == "bar"
    assert song.genre and song.genre.name == "Bar"


def test_caching_get_artists(cache_adapter: FilesystemAdapter):
    with pytest.raises(CacheMissError):
        cache_adapter.get_artists()

    # Ingest artists.
    cache_adapter.ingest_new_data(
        KEYS.ARTISTS,
        None,
        [
            SubsonicAPI.ArtistAndArtistInfo(
                id="1", name="test1", album_count=3, albums=[]
            ),
            SubsonicAPI.ArtistAndArtistInfo(id="2", name="test2", album_count=4),
        ],
    )

    artists = cache_adapter.get_artists()
    assert len(artists) == 2
    assert (artists[0].id, artists[0].name, artists[0].album_count) == ("1", "test1", 3)
    assert (artists[1].id, artists[1].name, artists[1].album_count) == ("2", "test2", 4)

    # Ingest a new artists list with one of them deleted.
    cache_adapter.ingest_new_data(
        KEYS.ARTISTS,
        None,
        [
            SubsonicAPI.ArtistAndArtistInfo(id="1", name="test1", album_count=3),
            SubsonicAPI.ArtistAndArtistInfo(id="3", name="test3", album_count=8),
        ],
    )

    # Now, artist 2 should be gone.
    artists = cache_adapter.get_artists()
    assert len(artists) == 2
    assert (artists[0].id, artists[0].name, artists[0].album_count) == ("1", "test1", 3)
    assert (artists[1].id, artists[1].name, artists[1].album_count) == ("3", "test3", 8)


def test_caching_get_ignored_articles(cache_adapter: FilesystemAdapter):
    with pytest.raises(CacheMissError):
        cache_adapter.get_ignored_articles()

    # Ingest ignored_articles.
    cache_adapter.ingest_new_data(KEYS.IGNORED_ARTICLES, None, {"Foo", "Bar"})
    artists = cache_adapter.get_ignored_articles()
    assert {"Foo", "Bar"} == artists

    # Ingest a new artists list with one of them deleted.
    cache_adapter.ingest_new_data(KEYS.IGNORED_ARTICLES, None, {"Foo", "Baz"})
    artists = cache_adapter.get_ignored_articles()
    assert {"Foo", "Baz"} == artists


def test_caching_get_artist(cache_adapter: FilesystemAdapter):
    with pytest.raises(CacheMissError):
        cache_adapter.get_artist("1")

    # Simulate the artist details being retrieved from Subsonic.
    cache_adapter.ingest_new_data(
        KEYS.ARTIST,
        "1",
        SubsonicAPI.ArtistAndArtistInfo(
            id="1",
            name="Bar",
            album_count=1,
            artist_image_url="image",
            similar_artists=[
                SubsonicAPI.ArtistAndArtistInfo(id="A", name="B"),
                SubsonicAPI.ArtistAndArtistInfo(id="C", name="D"),
            ],
            biography="this is a bio",
            music_brainz_id="mbid",
            albums=[
                SubsonicAPI.Album(id="1", name="Foo", _artist="Bar", artist_id="1")
            ],
        ),
    )

    artist = cache_adapter.get_artist("1")
    assert (
        artist.artist_image_url
        and (
            artist.id,
            artist.name,
            artist.album_count,
            artist.artist_image_url,
            artist.biography,
            artist.music_brainz_id,
        )
        == ("1", "Bar", 1, "image", "this is a bio", "mbid")
    )
    assert artist.similar_artists == [
        SubsonicAPI.ArtistAndArtistInfo(id="A", name="B"),
        SubsonicAPI.ArtistAndArtistInfo(id="C", name="D"),
    ]
    assert artist.albums and len(artist.albums) == 1
    assert cast(SelectQuery, artist.albums).dicts() == [
        SubsonicAPI.Album(id="1", name="Foo")
    ]

    # Simulate "force refreshing" the artist details being retrieved from Subsonic.
    cache_adapter.ingest_new_data(
        KEYS.ARTIST,
        "1",
        SubsonicAPI.ArtistAndArtistInfo(
            id="1",
            name="Foo",
            album_count=2,
            artist_image_url="image2",
            similar_artists=[
                SubsonicAPI.ArtistAndArtistInfo(id="A", name="B"),
                SubsonicAPI.ArtistAndArtistInfo(id="E", name="F"),
            ],
            biography="this is a bio2",
            music_brainz_id="mbid2",
            albums=[
                SubsonicAPI.Album(id="1", name="Foo", artist_id="1"),
                SubsonicAPI.Album(id="2", name="Bar", artist_id="1"),
            ],
        ),
    )

    artist = cache_adapter.get_artist("1")
    assert (
        artist.artist_image_url
        and (
            artist.id,
            artist.name,
            artist.album_count,
            artist.artist_image_url,
            artist.biography,
            artist.music_brainz_id,
        )
        == ("1", "Foo", 2, "image2", "this is a bio2", "mbid2")
    )
    assert artist.similar_artists == [
        SubsonicAPI.ArtistAndArtistInfo(id="A", name="B"),
        SubsonicAPI.ArtistAndArtistInfo(id="E", name="F"),
    ]
    assert artist.albums and len(artist.albums) == 2
    assert cast(SelectQuery, artist.albums).dicts() == [
        SubsonicAPI.Album(id="1", name="Foo"),
        SubsonicAPI.Album(id="2", name="Bar"),
    ]


def test_caching_get_album(cache_adapter: FilesystemAdapter):
    with pytest.raises(CacheMissError):
        cache_adapter.get_album("1")

    # Simulate the artist details being retrieved from Subsonic.
    cache_adapter.ingest_new_data(
        KEYS.ALBUM,
        "a1",
        SubsonicAPI.Album(
            id="a1",
            name="foo",
            cover_art="c",
            song_count=2,
            year=2020,
            duration=timedelta(seconds=31),
            play_count=20,
            _artist="cool",
            artist_id="art1",
            songs=MOCK_SUBSONIC_SONGS[:2],
        ),
    )

    album = cache_adapter.get_album("a1")
    assert album and album.cover_art
    assert (
        album.id,
        album.name,
        album.cover_art,
        album.song_count,
        album.year,
        album.play_count,
    ) == (
        "a1",
        "foo",
        "c",
        2,
        2020,
        20,
    )
    assert album.artist
    assert (album.artist.id, album.artist.name) == ("art1", "cool")
    assert album.songs
    verify_songs(album.songs, MOCK_SUBSONIC_SONGS[:2])


def test_caching_invalidate_artist(cache_adapter: FilesystemAdapter):
    # Simulate the artist details being retrieved from Subsonic.
    cache_adapter.ingest_new_data(
        KEYS.ARTIST,
        "artist1",
        SubsonicAPI.ArtistAndArtistInfo(
            id="artist1",
            name="Bar",
            album_count=1,
            artist_image_url="image",
            similar_artists=[
                SubsonicAPI.ArtistAndArtistInfo(id="A", name="B"),
                SubsonicAPI.ArtistAndArtistInfo(id="C", name="D"),
            ],
            biography="this is a bio",
            music_brainz_id="mbid",
            albums=[
                SubsonicAPI.Album(id="1", name="Foo", artist_id="1"),
                SubsonicAPI.Album(id="2", name="Bar", artist_id="1"),
            ],
        ),
    )
    cache_adapter.ingest_new_data(
        KEYS.ALBUM,
        "1",
        SubsonicAPI.Album(id="1", name="Foo", artist_id="artist1", cover_art="1"),
    )
    cache_adapter.ingest_new_data(
        KEYS.ALBUM,
        "2",
        SubsonicAPI.Album(id="2", name="Bar", artist_id="artist1", cover_art="2"),
    )
    cache_adapter.ingest_new_data(KEYS.COVER_ART_FILE, "image", MOCK_ALBUM_ART3)
    cache_adapter.ingest_new_data(KEYS.COVER_ART_FILE, "1", MOCK_ALBUM_ART)
    cache_adapter.ingest_new_data(KEYS.COVER_ART_FILE, "2", MOCK_ALBUM_ART2)

    stale_artist = cache_adapter.get_artist("artist1")
    stale_album_1 = cache_adapter.get_album("1")
    stale_album_2 = cache_adapter.get_album("2")
    stale_artist_artwork = cache_adapter.get_cover_art_uri("image", "file", size=300)
    stale_cover_art_1 = cache_adapter.get_cover_art_uri("1", "file", size=300)
    stale_cover_art_2 = cache_adapter.get_cover_art_uri("2", "file", size=300)

    cache_adapter.invalidate_data(KEYS.ARTIST, "artist1")

    # Test the cascade of cache invalidations.
    try:
        cache_adapter.get_artist("artist1")
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data
        assert e.partial_data == stale_artist

    try:
        cache_adapter.get_album("1")
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data
        assert e.partial_data == stale_album_1

    try:
        cache_adapter.get_album("2")
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data
        assert e.partial_data == stale_album_2

    try:
        cache_adapter.get_cover_art_uri("image", "file", size=300)
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data
        assert e.partial_data == stale_artist_artwork

    try:
        cache_adapter.get_cover_art_uri("1", "file", size=300)
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data
        assert e.partial_data == stale_cover_art_1

    try:
        cache_adapter.get_cover_art_uri("2", "file", size=300)
        assert 0, "DID NOT raise CacheMissError"
    except CacheMissError as e:
        assert e.partial_data
        assert e.partial_data == stale_cover_art_2


def test_get_music_directory(cache_adapter: FilesystemAdapter):
    dir_id = "d1"
    with pytest.raises(CacheMissError):
        cache_adapter.get_directory(dir_id)

    # Simulate the directory details being retrieved from Subsonic.
    cache_adapter.ingest_new_data(
        KEYS.DIRECTORY,
        dir_id,
        SubsonicAPI.Directory(
            dir_id,
            title="foo",
            parent_id=None,
            _children=[json.loads(s.to_json()) for s in MOCK_SUBSONIC_SONGS[:2]]
            + [
                {
                    "id": "542",
                    "parent": dir_id,
                    "isDir": True,
                    "title": "Crash My Party",
                }
            ],
        ),
    )

    directory = cache_adapter.get_directory(dir_id)
    assert directory and directory.id == dir_id
    assert directory.name == "foo"
    assert directory.parent_id == "root"

    dir_child, *song_children = directory.children
    verify_songs(song_children, MOCK_SUBSONIC_SONGS[:2])
    assert dir_child.id == "542"
    assert dir_child.parent_id
    assert dir_child.name == "Crash My Party"


def test_search(cache_adapter: FilesystemAdapter):
    with pytest.raises(CacheMissError):
        cache_adapter.get_artist("artist1")
    with pytest.raises(CacheMissError):
        cache_adapter.get_album("album1")
    with pytest.raises(CacheMissError):
        cache_adapter.get_song_details("s1")

    search_result = SublimeAPI.SearchResult("")
    search_result.add_results(
        "albums",
        [
            SubsonicAPI.Album(
                id="album1",
                name="Foo",
                artist_id="artist1",
                _artist="foo",
                cover_art="cal1",
            ),
            SubsonicAPI.Album(
                id="album2",
                name="Boo",
                artist_id="artist1",
                _artist="foo",
                cover_art="cal2",
            ),
        ],
    )
    search_result.add_results(
        "artists",
        [
            SubsonicAPI.ArtistAndArtistInfo(id="artist1", name="foo", cover_art="car1"),
            SubsonicAPI.ArtistAndArtistInfo(
                id="artist2", name="better boo", cover_art="car2"
            ),
        ],
    )
    search_result.add_results(
        "songs",
        [
            SubsonicAPI.Song(
                "s1",
                "amazing boo",
                cover_art="s1",
                _artist="artist3",
                artist_id="ohea1",
            ),
            SubsonicAPI.Song(
                "s2",
                "foo of all foo",
                cover_art="s2",
                _artist="artist4",
                artist_id="ohea2",
            ),
        ],
    )
    cache_adapter.ingest_new_data(KEYS.SEARCH_RESULTS, None, search_result)

    search_result = cache_adapter.search("foo")
    assert [
        (s.title, s.artist.name if s.artist else None) for s in search_result.songs
    ] == [("foo of all foo", "artist4"), ("amazing boo", "artist3")]
    assert [a.name for a in search_result.artists] == ["foo", "better boo"]
    assert [a.name for a in search_result.albums] == ["Foo", "Boo"]



import shutil
from pathlib import Path

import pytest

from sublime_music.adapters import ConfigurationStore
from sublime_music.adapters.filesystem import FilesystemAdapter
from sublime_music.adapters.subsonic import SubsonicAdapter
from sublime_music.config import AppConfiguration, ProviderConfiguration


@pytest.fixture
def config_filename(tmp_path: Path):
    yield tmp_path.joinpath("config.json")


@pytest.fixture
def cwd():
    yield Path(__file__).parent


def test_config_default_cache_location():
    config = AppConfiguration()
    assert config.cache_location == Path("~/.local/share/sublime-music").expanduser()


def test_server_property(tmp_path: Path):
    config = AppConfiguration()
    config.cache_location = tmp_path
    provider = ProviderConfiguration(
        id="1",
        name="foo",
        ground_truth_adapter_type=SubsonicAdapter,
        ground_truth_adapter_config=ConfigurationStore(),
    )
    config.providers["1"] = provider
    assert config.provider is None
    config.current_provider_id = "1"
    assert config.provider == provider

    assert config._state_file_location == tmp_path.joinpath("1", "state.pickle")


def test_json_load_unload(config_filename: Path, tmp_path: Path):
    ConfigurationStore.MOCK = True
    subsonic_config_store = ConfigurationStore(username="test")
    subsonic_config_store.set_secret("password", "testpass")
    original_config = AppConfiguration(
        providers={
            "1": ProviderConfiguration(
                id="1",
                name="foo",
                ground_truth_adapter_type=SubsonicAdapter,
                ground_truth_adapter_config=subsonic_config_store,
                caching_adapter_type=FilesystemAdapter,
                caching_adapter_config=ConfigurationStore(),
            )
        },
        current_provider_id="1",
        filename=config_filename,
    )
    original_config.cache_location = tmp_path

    original_config.save()

    loaded_config = AppConfiguration.load_from_file(config_filename)

    assert original_config.version == loaded_config.version
    assert original_config.providers == loaded_config.providers
    assert original_config.provider == loaded_config.provider


def test_config_migrate_v5_to_v6(config_filename: Path, cwd: Path):
    shutil.copyfile(str(cwd.joinpath("mock_data/config-v5.json")), str(config_filename))
    app_config = AppConfiguration.load_from_file(config_filename)
    app_config.migrate()

    assert app_config.version == 6
    assert app_config.player_config == {
        "Local Playback": {"Replay Gain": "track"},
        "Chromecast": {
            "Serve Local Files to Chromecasts on the LAN": True,
            "LAN Server Port Number": 6969,
        },
    }
    app_config.save()
    app_config2 = AppConfiguration.load_from_file(config_filename)
    assert app_config == app_config2
